__all__ = ('GuildJoinRequestDeleteEvent',)

from scarletio import copy_docs

from ...bases import EventBase
from ...core import GUILDS
from ...user import create_partial_user_from_id

from .fields import (
    parse_guild_id, parse_user_id, put_guild_id, put_user_id, validate_guild_id, validate_user_id
)


class GuildJoinRequestDeleteEvent(EventBase):
    """
    Represents a `GUILD_JOIN_REQUEST_DELETE` event.
    
    Attributes
    ----------
    guild_id : `int`
        The guild's identifier where the event is for.
    
    user_id : `int`
        The user's identifier whos join request was deleted.
    """
    __slots__ = ('guild_id', 'user_id', )
    
    def __new__(cls, *, guild_id = ..., user_id = ...):
        """
        Creates a new guild join request delete event.
        
        Parameters
        ----------
        guild_id : ``None | int | Guild``, Optional (Keyword only)
            The guild or its identifier where the event will be.
        
        user_id : ``None | int | ClientUserBase``, Optional (Keyword only)
            The user or it's identifier whos join request was deleted.
        
        Raises
        ------
        TypeError
            - If a parameter's type is invalid.
        ValueError
            - If a parameter's value is invalid.
        """
        # guild_id
        if guild_id is ...:
            guild_id = 0
        else:
            guild_id = validate_guild_id(guild_id)
        
        # user_id
        if user_id is ...:
            user_id = 0
        else:
            user_id = validate_user_id(user_id)
        
        # Construct
        self = object.__new__(cls)
        self.guild_id = guild_id
        self.user_id = user_id
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new guild join request delete event from the given data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Guild join request delete event data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.guild_id = parse_guild_id(data)
        self.user_id = parse_user_id(data)
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Serializes the guild join request delete event.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether fields of their default value should be included as well.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        put_guild_id(self.guild_id, data, defaults)
        put_user_id(self.user_id, data, defaults)
        return data
    
    
    @copy_docs(EventBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', type(self).__name__]
        
        # guild_id
        repr_parts.append(' guild_id = ')
        repr_parts.append(repr(self.guild_id))
        
        # user_id
        repr_parts.append(', user_id = ')
        repr_parts.append(repr(self.user_id))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(EventBase.__len__)
    def __len__(self):
        return 2
    
    
    @copy_docs(EventBase.__iter__)
    def __iter__(self):
        yield self.guild_id
        yield self.user_id
    
    
    @copy_docs(EventBase.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        # guild_id
        if self.guild_id != other.guild_id:
            return False
        
        # user_id
        if self.user_id != other.user_id:
            return False
        
        return True
    
    
    @copy_docs(EventBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # guild_id
        hash_value ^= self.guild_id
        
        # user_id
        hash_value ^= self.user_id
        
        return hash_value
    
    
    def copy(self):
        """
        Copies the guild join request delete event.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.guild_id = self.guild_id
        new.user_id = self.user_id
        return new
    
    
    def copy_with(self, *, guild_id = ..., user_id = ...):
        """
        Copies the guild join request delete event with the given fields.
        
        Parameters
        ----------
        guild_id : ``None | int | Guild``, Optional (Keyword only)
            The guild or its identifier where the event will be.
        
        user_id : ``None | int | ClientUserBase``, Optional (Keyword only)
            The user or it's identifier whos join request was deleted.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is invalid.
        ValueError
            - If a parameter's value is invalid.
        """
        # guild_id
        if guild_id is ...:
            guild_id = self.guild_id
        else:
            guild_id = validate_guild_id(guild_id)
        
        # user_id
        if user_id is ...:
            user_id = self.user_id
        else:
            user_id = validate_user_id(user_id)
        
        # Construct
        new = object.__new__(type(self))
        new.guild_id = guild_id
        new.user_id = user_id
        return new
    
    
    @property
    def guild(self):
        """
        Returns the guild join request delete event's guild.
        
        Returns
        -------
        guild : ``None | Guild``
        """
        guild_id = self.guild_id
        if guild_id:
            return GUILDS.get(guild_id, None)
    
    
    @property
    def user(self):
        """
        Returns the user whos join request was deleted.
        
        Returns
        -------
        user : ``ClientUserBase``
        """
        return create_partial_user_from_id(self.user_id)



__all__ = ()

from importlib.machinery import PathFinder, ExtensionFileLoader

from ..constants import PLUGINS
from ..plugin_root import is_in_plugin_root

from .spec_finder_helpers import find_spec_in_paths, is_spec_in_test_directory
from .module_spec_type import PluginModuleSpecType


class PluginFinder(PathFinder):
    """
    Plugin finder type. Subclass of ``PathFinder``
    """
    @classmethod
    def find_spec(cls, full_name, paths = None, target = None):
        """
        Tries to find the module specification of the given name.
        
        Parameters
        ----------
        full_name : `str`
            The full name of the module to find.
        paths : `None | list<str>` = `None`, Optional
            Path to find the module in.
        target : `None | object` = `None`, Optional
            Helper value to find the target specification.
            
            > Could not find a case where this value is actually used.
        
        Returns
        -------
        module_specification : `None`, ``PluginModuleSpecType``
        """
        try:
            plugin = PLUGINS[full_name]
        except KeyError:
            pass
        else:
            return plugin._spec
        
        if not is_in_plugin_root(full_name):
            return None
        
        spec = find_spec_in_paths(full_name, paths)
        if spec is None:
            return None
        
        if is_spec_in_test_directory(spec):
            return None
        
        loader = spec.loader
        if loader is None:
            return None
        
        if isinstance(loader, ExtensionFileLoader):
            return spec
        
        return PluginModuleSpecType(spec)



__all__ = ()

from ...field_parsers import entity_id_array_parser_factory
from ...field_putters import nullable_entity_id_array_putter_factory, int_optional_putter_factory
from ...field_validators import (
    entity_id_array_validator_factory, entity_id_set_validator_factory, int_conditional_validator_factory
)
from ...user import ClientUserBase

from .constants import DELETE_MESSAGE_DURATION_DEFAULT, DELETE_MESSAGE_DURATION_MIN, DELETE_MESSAGE_DURATION_MAX

# banned_user_ids

parse_banned_user_ids = entity_id_array_parser_factory('banned_users')
put_banned_user_ids = nullable_entity_id_array_putter_factory('banned_users')
validate_banned_user_ids = entity_id_array_validator_factory('banned_user_ids', ClientUserBase)

# delete_message_duration

put_delete_message_duration = int_optional_putter_factory(
    'delete_message_seconds', DELETE_MESSAGE_DURATION_DEFAULT
)

validate_delete_message_duration = int_conditional_validator_factory(
    'delete_message_duration',
    DELETE_MESSAGE_DURATION_DEFAULT,
    (
        lambda delete_message_duration :
            delete_message_duration >= DELETE_MESSAGE_DURATION_MIN and
            delete_message_duration <= DELETE_MESSAGE_DURATION_MAX
    ),
    f'>= {DELETE_MESSAGE_DURATION_MIN!r} and <= {DELETE_MESSAGE_DURATION_MAX!r}',
)


# failed_user_ids

parse_failed_user_ids = entity_id_array_parser_factory('failed_users')
put_failed_user_ids = nullable_entity_id_array_putter_factory('failed_users')
validate_failed_user_ids = entity_id_array_validator_factory('failed_user_ids', ClientUserBase)

# user_ids

def put_user_ids(user_ids, data, defaults):
    """
    Puts the user identifiers into the given data.
    
    Parameters
    ----------
    user_ids : `set<int>`
        The user identifiers to put into the given data.
    data : `dict<str, object>`
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    data['user_ids'] = [str(user_id) for user_id in user_ids]
    return data


validate_user_ids = entity_id_set_validator_factory('user_ids', ClientUserBase)




__all__ = ()

from ...field_parsers import preinstanced_parser_factory
from ...field_putters import preinstanced_putter_factory
from ...field_validators import preinstanced_validator_factory

from .preinstanced import ComponentType


# type

parse_type = preinstanced_parser_factory('type', ComponentType, ComponentType.none)
put_type = preinstanced_putter_factory('type')
validate_type = preinstanced_validator_factory('component_type', ComponentType)



__all__ = ()

from scarletio import RichAttributeErrorBaseType

from .array import ArrayUInt32BE
from .constants import MAX_UINT_16, MAX_UINT_32


# used: 
# http://www.rfcreader.com/#rfc3550_line548
#
# not used:
# http://www.rfcreader.com/#rfc3550_line855
# http://www.rfcreader.com/#rfc3550_line1614
# #http://www.rfcreader.com/#rfc3550_line1879

# http://www.rfcreader.com/#rfc3550_line548

#    0                   1                   2                   3
#    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
#   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#   |V=2|P|E|  CC   |M|     PT      |       sequence number         |
#   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#   |                           timestamp                           |
#   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#   |                             source                            |
#   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#
# For each `CC` (contributing source count) :
#   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#   |               contributing source identifiers                 |
#   |                             ....                              |
#   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#
# If `E` (extended)
#   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#   |       extension profile       |      extension count          |
#   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#
# For each `extension count`
#   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#   |                       header extensions                       |
#   |                             ....                              |
#   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#
#   +=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+
#   |                           payload                             |
#   +=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+
#
# For each `padding count - 1`, because it also includes the padding count field:
#   +-+-+-+-+-+-+-+-+
#   |    padding    |
#   +-+-+-+-+-+-+-+-+
#
# If `P` (padded)
#   +-+-+-+-+-+-+-+-+
#   | padding count |
#   +-+-+-+-+-+-+-+-+

def _iter_contributing_source_bytes(contributing_sources):
    """
    Iterates over and converts the given contributing sources to bytes.
    
    This function is an iterable generator.
    
    Parameters
    ----------
    contributing_sources : `None | sequence<int>`
        Contributing sources.
    
    Yields
    ------
    source_bytes : `bytes`
    """
    if (contributing_sources is not None):
        for contributing_source in contributing_sources:
            yield contributing_source.to_bytes(4, 'big')


def _iter_extension_bytes(extension_profile, extensions):
    """
    Iterates over and converts the given extension information to bytes.
    
    This function is an iterable generator.
    
    Parameters
    ----------
    extension_profile : `int`
        Extension profile.
    extensions : `None | sequence<int>`
        Extensions.
    
    Yields
    ------
    extension_bytes : `bytes`
    """
    if extension_profile or (extensions is not None):
        yield extension_profile.to_bytes(2, 'big')
        yield (0 if extensions is None else len(extensions)).to_bytes(2, 'big')
        
        if (extensions is not None):
            for extension in extensions:
                yield extension.to_bytes(4, 'big')


def _iter_padding_bytes(padding):
    """
    Yields back the padding. 
    
    This function is an iterable generator.
    
    Parameters
    ----------
    padding : `None | bytes-like`
        Padding if any.
    
    Yields
    ------
    padding_bytes : `bytes-like`
    """
    if (padding is not None):
        yield padding


class RTPPacket(RichAttributeErrorBaseType):
    """
    Represents an RTP packet.
    
    Attributes
    ----------
    data : `bytes`
        The raw data of the packet on what it is casted on.
    """
    __slots__ = ('data',)
    
    def __new__(cls, data):
        """
        Creates a new rtp packet with the given data.
        
        Parameters
        ----------
        data : `bytes`
        """
        self = object.__new__(cls)
        self.data = data
        return self
    
    
    @property
    def version(self):
        """
        Returns the rtp packet's version.
        
        Returns
        -------
        version : `int`
        """
        return self.data[0] >> 6
    
    
    @property
    def padded(self):
        """
        Returns whether the rtp packet is padded.
        
        Returns
        -------
        padded : `bool`
        """
        return True if (self.data[0] & 0b00100000) else False
    
    
    @property
    def extended(self):
        """
        Returns whether the rtp packet is extended.
        
        Returns
        -------
        extended : `bool`
        """
        return True if (self.data[0] & 0b00010000) else False
    
    
    @property
    def contributing_source_count(self):
        """
        Returns the contributing source count.
        
        Returns
        -------
        contributing_source_count : `int`
        """
        return self.data[0] & 0b00001111
    
    
    @property
    def marker(self):
        """
        Returns whether the packet is a marker.
        
        Returns
        -------
        marker : `bool`
        """
        return True if (self.data[1] & 0b10000000) else False
    
    
    @property
    def payload_type(self):
        """
        Returns the payload's type.
        
        Returns
        -------
        payload_type : `int`
        """
        return self.data[1] & 0b01111111
    
    
    @property
    def sequence(self):
        """
        Returns the sequence number.
        
        Returns
        -------
        sequence : `bool`
        """
        return int.from_bytes(self.data[2 : 4], 'big')
    
    
    @property
    def timestamp(self):
        """
        Returns the timestamp.
        
        Returns
        -------
        timestamp : `bool`
        """
        return int.from_bytes(self.data[4 : 8], 'big')
    
    
    @property
    def source(self):
        """
        Returns the source number.
        
        Returns
        -------
        source : `int`
        """
        return int.from_bytes(self.data[8 : 12], 'big')
    
    
    @property
    def header(self):
        """
        Returns the base of the header.
        
        Returns
        -------
        header : `bytes | memoryview`
        """
        data = self.data
        if len(data) <= 12:
            return data
        
        return memoryview(self.data)[ : 12]
    
    
    @property
    def contributing_sources(self):
        """
        Returns the contributing sources.
        
        Returns
        -------
        contributing_sources : `None | ArrayUInt32BE`
        """
        if not self.contributing_source_count:
            return None
        
        return ArrayUInt32BE(self.data, 12, self.extension_header_start)
    
    
    @property
    def extension_header_start(self):
        """
        Returns where the extension headers start.
        
        Returns
        -------
        extension_header_start : `int`
        """
        return 12 + (self.contributing_source_count << 2)
    
    
    @property
    def extension_profile(self):
        """
        Returns the extension profile.
        
        Returns
        -------
        extension_profile : `int`
        """
        if not self.extended:
            return 0
        
        start = self.extension_header_start
        return int.from_bytes(self.data[start : start + 2], 'big')
    
    
    @property
    def extension_count(self):
        """
        Returns the amount of extensions.
        
        Returns
        -------
        extension_count : `int`
        """
        if not self.extended:
            return 0
        
        start = self.extension_header_start + 2
        length = int.from_bytes(self.data[start : start + 2], 'big')
        return length
    
    
    @property
    def extensions(self):
        """
        Returns the extension values.
        
        Returns
        -------
        extensions : `None | ArrayUInt32BE`
        """
        extension_count = self.extension_count
        if not extension_count:
            return None
        
        
        start = self.extension_header_start + 4
        end = start + (extension_count << 2)
        
        return ArrayUInt32BE(self.data, start, end)
    
    
    @property
    def payload_start(self):
        """
        Returns where the payload starts.
        
        Returns
        -------
        payload_start : `int`
        """
        start = self.extension_header_start
        extension_count = self.extension_count
        if extension_count:
            start += 4 + (extension_count << 2)
        
        return start
    
    
    @property
    def payload_end(self):
        """
        Returns where the payload ends.
        
        Returns
        -------
        payload_end : `int`
        """
        return len(self.data) - self.padding_count
    
    
    @property
    def payload(self):
        """
        Returns the payload.
        
        Returns
        -------
        payload : `memoryview`
        """
        return memoryview(self.data)[self.payload_start : self.payload_end]
    
    
    @property
    def padding_count(self):
        """
        Returns the amount of paddings.
        
        Returns
        -------
        padding_count : `int`
        """
        padded = self.padded
        if not padded:
            return 0
        
        return self.data[-1]
    
    
    @property
    def padding(self):
        """
        Returns the padding.
        
        Returns
        -------
        padding : `None | memoryview`
        """
        padding_count = self.padding_count
        if not padding_count:
            return None
        
        return memoryview(self.data)[-padding_count : ]
    
    
    @classmethod
    def from_fields(
        cls,
        version,
        padding,
        extension_profile,
        extensions,
        contributing_sources,
        marker,
        payload_type,
        sequence_number,
        timestamp,
        source,
        payload,
    ):
        """
        Creates a new rtp packet from the given fields.
        
        Parameters
        ----------
        version : `int`
            Version number.
        
        padding : `None | bytes-like`
            Padding if any.
        
        extension_profile : `int`
            Extension profile.
        
        extensions : `None | sequence<int>`
            Extensions.
        
        contributing_sources : `None | sequence<int>`
            Contributing sources.
        
        marker : `bool`
            Whether the packet is a marker.
        
        payload_type : `int`
            The payload's type.
        
        sequence_number : `int`
            Sequence number.
        
        timestamp : `int`
            Timestamp.
        
        source : `int`
            Packet source.
        
        payload : `bytes-like`
            The payload itself.
        
        Returns
        -------
        self : `instance<cls>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is invalid.
        ValueError
            - If a parameter's value is invalid.
        """
        if extension_profile or (extensions is not None):
            extended = True
        else:
            extended = False
        
        if (padding is not None) or (not padding):
            padded = True
            if padding[-1] != len(padding):
                raise ValueError(f'Last byte of batting must be equal to its length; got padding = {padding!r}')
        
        else:
            padded = False
        
        if contributing_sources is None:
            contributing_source_count = 0
        else:
            contributing_source_count = len(contributing_sources)
        
        
        data = b''.join([
            ((version << 6) | ((padded << 5)) | (extended << 4) | contributing_source_count).to_bytes(1, 'big'),
            ((marker << 7) | payload_type).to_bytes(1, 'big'),
            (sequence_number & MAX_UINT_16).to_bytes(2, 'big'),
            (timestamp & MAX_UINT_32).to_bytes(4, 'big'),
            source.to_bytes(4, 'big'),
            *_iter_contributing_source_bytes(contributing_sources),
            *_iter_extension_bytes(extension_profile, extensions),
            payload,
            *_iter_padding_bytes(padding),
        ])
        
        self = object.__new__(cls)
        self.data = data
        return self
    
    
    def __repr__(self):
        """Returns the packet's representation."""
        repr_parts = ['<', type(self).__name__]
        
        # version
        repr_parts.append(' version = ')
        repr_parts.append(repr(self.version))
        
        # padded (padding)
        padding = self.padding
        if (padding is not None):
            repr_parts.append(', padding = ')
            repr_parts.append(repr(padding))
        
        # extended (extension_profile, extension)
        extension_profile = self.extension_profile
        extensions = self.extensions
        if extension_profile or (extensions is not None):
            repr_parts.append(', extension_profile = ')
            repr_parts.append(repr(extension_profile))
            
            repr_parts.append(', extensions = ')
            repr_parts.append(repr(extensions))
        
        # contributing_sources
        contributing_sources = self.contributing_sources
        if (contributing_sources is not None):
            repr_parts.append(', contributing_sources = ')
            repr_parts.append(repr(contributing_sources))
        
        # marker
        marker = self.marker
        if marker:
            repr_parts.append(', marker = ')
            repr_parts.append(repr(marker))
        
        # payload_type
        repr_parts.append(', payload_type = ')
        repr_parts.append(repr(self.payload_type))
        
        # sequence
        repr_parts.append(', sequence = ')
        repr_parts.append(repr(self.sequence))
        
        # timestamp
        repr_parts.append(', timestamp = ')
        repr_parts.append(repr(self.timestamp))
        
        # source
        repr_parts.append(', source = ')
        repr_parts.append(repr(self.source))
        
        # payload
        payload = self.payload
        if payload:
            repr_parts.append(', payload (length) = ')
            repr_parts.append(repr(len(payload)))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two rtp packets are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self.data == other.data
    
    
    def __hash__(self):
        """Returns the rtp packet's hash value."""
        return hash(self.data)



__all__ = ()

from ...component import ComponentType
from ...field_putters import preinstanced_putter_factory
from ...field_validators import preinstanced_validator_factory


# type

def parse_type(data):
    """
    Parses component type out from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    component_type : ``ComponentType``
    """
    for key in ('type', 'component_type'):
        component_type_raw = data.get(key, None)
        if (component_type_raw is not None):
            return ComponentType(component_type_raw)
    
    return ComponentType.none


put_type = preinstanced_putter_factory('type')
validate_type = preinstanced_validator_factory('type', ComponentType)



__all__ = (
    'create_interaction_guild_data', 'create_partial_guild_data', 'create_partial_guild_from_data',
    'create_partial_guild_from_id', 'create_partial_guild_from_interaction_guild_data'
)

from functools import partial as partial_func

from scarletio import export

from ...core import GUILDS

from .fields import (
    parse_available, parse_description, parse_features, parse_id, parse_locale, parse_name, parse_nsfw_level,
    parse_verification_level, put_afk_channel_id, put_afk_timeout, put_available,
    put_boost_progress_bar_enabled, put_channels_and_channel_datas,
    put_default_message_notification_level, put_description, put_explicit_content_filter_level,
    put_features, put_hub_type, put_id, put_locale, put_mfa_level, put_name,
    put_nsfw_level, put_owner_id, put_public_updates_channel_id, put_roles_and_role_datas,
    put_rules_channel_id, put_safety_alerts_channel_id, put_system_channel_flags,
    put_system_channel_id, put_vanity_code, put_verification_level, put_widget_channel_id,
    put_widget_enabled, validate_afk_channel_id, validate_afk_timeout, validate_boost_progress_bar_enabled,
    validate_channels_and_channel_datas, validate_default_message_notification_level, validate_description,
    validate_explicit_content_filter_level, validate_features, validate_hub_type, validate_locale, validate_mfa_level,
    validate_name, validate_nsfw_level, validate_owner_id, validate_public_updates_channel_id,
    validate_roles_and_role_datas, validate_rules_channel_id, validate_safety_alerts_channel_id,
    validate_system_channel_flags, validate_system_channel_id, validate_vanity_code, validate_verification_level,
    validate_widget_channel_id, validate_widget_enabled
)
from .flags import SystemChannelFlag
from .guild import GUILD_BANNER, GUILD_DISCOVERY_SPLASH, GUILD_ICON, GUILD_INVITE_SPLASH, Guild
from .preinstanced import ExplicitContentFilterLevel, MessageNotificationLevel, NsfwLevel, VerificationLevel


GUILD_FIELD_CONVERTERS = {
    'afk_channel_id': (validate_afk_channel_id, put_afk_channel_id),
    'afk_timeout': (validate_afk_timeout, put_afk_timeout),
    'banner': (
        partial_func(GUILD_BANNER.validate_icon, allow_data = True),
        partial_func(GUILD_BANNER.put_into, as_data = True),
    ),
    'boost_progress_bar_enabled': (validate_boost_progress_bar_enabled, put_boost_progress_bar_enabled),
    'default_message_notification_level': (
        validate_default_message_notification_level, put_default_message_notification_level)
    ,
    'description': (validate_description, put_description),
    'discovery_splash': (
        partial_func(GUILD_DISCOVERY_SPLASH.validate_icon, allow_data = True),
        partial_func(GUILD_DISCOVERY_SPLASH.put_into, as_data = True),
    ),
    'explicit_content_filter_level': (validate_explicit_content_filter_level, put_explicit_content_filter_level),
    'features': (validate_features, put_features),
    'hub_type': (validate_hub_type, put_hub_type),
    'icon': (
        partial_func(GUILD_ICON.validate_icon, allow_data = True),
        partial_func(GUILD_ICON.put_into, as_data = True),
    ),
    'invite_splash': (
        partial_func(GUILD_INVITE_SPLASH.validate_icon, allow_data = True),
        partial_func(GUILD_INVITE_SPLASH.put_into, as_data = True),
    ),
    'locale': (validate_locale, put_locale),
    'mfa_level': (validate_mfa_level, put_mfa_level),
    'name': (validate_name, put_name),
    'nsfw_level': (validate_nsfw_level, put_nsfw_level),
    'owner_id': (validate_owner_id, put_owner_id),
    'public_updates_channel_id': (validate_public_updates_channel_id, put_public_updates_channel_id),
    'rules_channel_id': (validate_rules_channel_id, put_rules_channel_id),
    'safety_alerts_channel_id': (validate_safety_alerts_channel_id, put_safety_alerts_channel_id),
    'system_channel_flags': (validate_system_channel_flags, put_system_channel_flags),
    'system_channel_id': (validate_system_channel_id, put_system_channel_id),
    'vanity_code': (validate_vanity_code, put_vanity_code),
    'verification_level': (validate_verification_level, put_verification_level),
    'widget_channel_id': (validate_widget_channel_id, put_widget_channel_id),
    'widget_enabled': (validate_widget_enabled, put_widget_enabled),
}


# We need to ignore client adding, because clients count to being not partial.
# If a guild is not partial it wont get update on Guild.__new__

def create_partial_guild_from_data(data):
    """
    Creates a partial guild from partial guild data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Partial channel data received from Discord.
    
    Returns
    -------
    guild : ``Guild``
    """
    guild_id = parse_id(data)
    try:
        return GUILDS[guild_id]
    except KeyError:
        pass
    
    guild = Guild._create_empty(guild_id)
    
    guild.available = parse_available(data)
    guild.description = parse_description(data)
    guild._set_banner(data)
    guild._set_discovery_splash(data)
    guild._set_icon(data)
    guild._set_invite_splash(data)
    guild.name = parse_name(data)
    
    # Optional
    features = parse_features(data)
    if (features is not None):
        guild.features = features
    
    nsfw_level = parse_nsfw_level(data)
    if nsfw_level is not NsfwLevel.none:
        guild.nsfw_level = nsfw_level
    
    verification_level = parse_verification_level(data)
    if verification_level is not VerificationLevel.none:
        guild.verification_level = verification_level
        
    GUILDS[guild_id] = guild
    
    return guild


def create_partial_guild_data(guild):
    """
    Creates partial guild data. The opposite of ``create_partial_guild_from_data``.
    
    Parameters
    ----------
    guild : ``Guild``
        The guild to serialize.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    data = {}
    
    put_available(guild.available, data, True)
    put_description(guild.description, data, True)
    type(guild).discovery_splash.put_into(guild.discovery_splash, data, True, as_data = False)
    put_features(guild.features, data, True)
    put_id(guild.id, data, True)
    type(guild).icon.put_into(guild.icon, data, True, as_data = False)
    type(guild).invite_splash.put_into(guild.invite_splash, data, True, as_data = False)
    put_name(guild.name, data, True)
    put_nsfw_level(guild.nsfw_level, data, True)
    put_verification_level(guild.verification_level, data, True)
    
    return data


@export
def create_partial_guild_from_id(guild_id):
    """
    Creates a guild from the given identifier and stores it in the cache as well. If the guild already exists,
    returns that instead.
    
    Parameters
    ----------
    guild_id : `int`
        The respective guild's identifier.
    
    Returns
    -------
    guild : ``Guild``
        The created guild instance.
    """
    try:
        guild = GUILDS[guild_id]
    except KeyError:
        guild = Guild._create_empty(guild_id)
        GUILDS[guild_id] = guild
    
    return guild


def create_new_guild_data(
    *,
    afk_channel_id = ...,
    afk_timeout = ...,
    channels = ...,
    default_message_notification_level = ...,
    explicit_content_filter_level = ...,
    icon = ...,
    name = ...,
    roles = ...,
    system_channel_flags = ...,
    system_channel_id = ...,
    verification_level = ...,
):
    """
    Helper function to create guild create data.
    
    Parameters
    ----------
    afk_channel_id : `None | int`, Optional (Keyword only)
        The id of the guild's afk channel. The id should be one of the channel's id from `channels`.
    
    afk_timeout : `None | int` = `None`, Optional (Keyword only)
        The afk timeout for the users at the guild's afk channel.
    
    channels : `None`, `list` of `dict`, Optional (Keyword only)
        A list of channels of the new guild. It should contain channel data objects.
    
    default_message_notification_level : ``MessageNotificationLevel``, `int`, Optional (Keyword only)
        The message notification level of the new guild.
    
    explicit_content_filter_level : ``ExplicitContentFilterLevel``, `int`, Optional (Keyword only)
        The content filter level of the guild.
    
    icon : `None`, `bytes-like`, Optional (Keyword only)
        The icon of the new guild.
    
    name : `str`, Optional (Keyword only)
        The name of the new guild.
    
    roles : `None`, `list` of (`dict<str, object>`, ``Role``), Optional (Keyword only)
        A list of roles of the new guild. It should contain role data objects.
    
    system_channel_flags : ``SystemChannelFlag``, `int`, Optional (Keyword only)
        Describe which type of messages are sent automatically to the system channel.
    
    system_channel_id: `int`, Optional (Keyword only)
        The id of the guild's system channel. The id should be one of the channel's id from `channels`.
    
    verification_level : ``VerificationLevel``, `int`, Optional (Keyword only)
        The verification level of the new guild.
    
    Returns
    -------
    data : `dict<str, object>`
    
    Raises
    ------
    TypeError
        - If a parameter's type is incorrect.
    ValueError
        - If a parameter's value is incorrect.
    """
    # afk_channel_id
    if afk_channel_id is ...:
        afk_channel_id = 0
    else:
        afk_channel_id = validate_afk_channel_id(afk_channel_id)
    
    # afk_timeout
    if afk_timeout is ...:
        afk_timeout = 0
    else:
        afk_timeout = validate_afk_timeout(afk_timeout)
    
    # channels
    if channels is ...:
        channels = None
    else:
        channels = validate_channels_and_channel_datas(channels)
    
    # default_message_notification_level
    if default_message_notification_level is ...:
        default_message_notification_level = MessageNotificationLevel.all_messages
    else:
        default_message_notification_level = validate_default_message_notification_level(default_message_notification_level)
    
    # explicit_content_filter_level
    if explicit_content_filter_level is ...:
        explicit_content_filter_level = ExplicitContentFilterLevel.disabled
    else:
        explicit_content_filter_level = validate_explicit_content_filter_level(explicit_content_filter_level)
    
    if icon is ...:
        icon = None
    else:
        icon = GUILD_ICON.validate_icon(icon, allow_data = True)
    
    # name
    if name is ...:
        name = ''
    else:
        name = validate_name(name)
    
    # roles
    if roles is ...:
        roles = None
    else:
        roles = validate_roles_and_role_datas(roles)
    
    # system_channel_id
    if system_channel_id is ...:
        system_channel_id = 0
    else:
        system_channel_id = validate_system_channel_id(system_channel_id)
    
    # system_channel_flags
    if system_channel_flags is ...:
        system_channel_flags = SystemChannelFlag.NONE
    else:
        system_channel_flags = validate_system_channel_flags(system_channel_flags)
    
    # verification_level
    if verification_level is ...:
        verification_level = VerificationLevel.none
    else:
        verification_level = validate_verification_level(verification_level)
    
    data = {}
    put_name(name, data, True)
    put_afk_channel_id(afk_channel_id, data, True)
    put_afk_timeout(afk_timeout, data, False)
    put_channels_and_channel_datas(channels, data, True)
    put_explicit_content_filter_level(explicit_content_filter_level, data, True)
    GUILD_ICON.put_into(icon, data, defaults = True, as_data = True)
    put_roles_and_role_datas(roles, data,  True)
    put_default_message_notification_level(default_message_notification_level, data, True)
    put_system_channel_id(system_channel_id, data, True)
    put_system_channel_flags(system_channel_flags, data, True)
    put_verification_level(verification_level, data, True)
    return data


def create_partial_guild_from_interaction_guild_data(data):
    """
    Creates a partial guild from the guild data received through an interaction.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Interaction guild data.
    
    Returns
    -------
    guild : ``Guild``
    """
    guild_id = parse_id(data)
    try:
        guild = GUILDS[guild_id]
    except KeyError:
        guild = Guild._create_empty(guild_id)
        GUILDS[guild_id] = guild
    
    guild.locale = parse_locale(data)
    guild.features = parse_features(data)
    
    return guild


def create_interaction_guild_data(guild):
    """
    Creates interaction guild data.
    
    Parameters
    ----------
    guild : ``Guild``
        The guild to serialise.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    data = {}
    
    put_id(guild.id, data, True)
    put_locale(guild.locale, data, True)
    put_features(guild.features, data, True)
    
    return data



__all__ = ('SharedClientThemeBaseTheme',)

from ...bases import Preinstance as P, PreinstancedBase


class SharedClientThemeBaseTheme(PreinstancedBase, value_type = int):
    """
    Represents an shared theme's base type.
    
    Attributes
    ----------
    name : `str`
        The name of the type.
    
    value : `int`
        The Discord side identifier value of the type.
        
    Type Attributes
    ---------------
    Every predefined shared theme base type can be accessed as type attribute as well:
    
    +-----------------------+---------------+-------+
    | Type attribute name   | name          | Value |
    +=======================+===============+=======+
    | none                  | none          | 0     |
    +-----------------------+---------------+-------+
    | dark                  | dark          | 1     |
    +-----------------------+---------------+-------+
    | light                 | light         | 2     |
    +-----------------------+---------------+-------+
    | darker                | darker        | 3     |
    +-----------------------+---------------+-------+
    | midnight              | midnight      | 4     |
    +-----------------------+---------------+-------+
    """
    __slots__ = ()
    
    # predefined
    none = P(0, 'none')
    dark = P(1, 'dark')
    light = P(2, 'light')
    darker = P(3, 'darker')
    midnight = P(4, 'midnight')



__all__ = ('RoleManagerMetadataIntegration',)

from scarletio import copy_docs, include

from .base import RoleManagerMetadataBase
from .fields import parse_integration_id, put_integration_id, validate_integration_id


create_partial_integration_from_id = include('create_partial_integration_from_id')


class RoleManagerMetadataIntegration(RoleManagerMetadataBase):
    """
    Role manager metadata of a role managed by a integration.
    
    Attributes
    ----------
    integration_id : `int`
        The manager integration's identifier.
    """
    __slots__ = ('integration_id', )
    
    def __new__(cls, *, integration_id = ...):
        """
        Creates a new role manager.
        
        Parameters
        ----------
        integration_id : `int`, ``Integration``, Optional (Keyword only)
            The manager integration's identifier.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # integration_id
        if integration_id is ...:
            integration_id = 0
        else:
            integration_id = validate_integration_id(integration_id)
        
        self = object.__new__(cls)
        self.integration_id = integration_id
        return self
    
    
    @classmethod
    @copy_docs(RoleManagerMetadataBase.from_data)
    def from_data(cls, data):
        self = object.__new__(cls)
        self.integration_id = parse_integration_id(data)
        return self
    
    
    @copy_docs(RoleManagerMetadataBase.to_data)
    def to_data(self, *, defaults = False):
        data = {}
        put_integration_id(self.integration_id, data, defaults)
        return data
    
    
    @copy_docs(RoleManagerMetadataBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' integration_id = ')
        repr_parts.append(repr(self.integration_id))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(RoleManagerMetadataBase.__hash__)
    def __hash__(self):
        return self.integration_id
    
    
    @copy_docs(RoleManagerMetadataBase.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        return self.integration_id == other.integration_id
    
    
    @copy_docs(RoleManagerMetadataBase.copy)
    def copy(self):
        new = object.__new__(type(self))
        new.integration_id = self.integration_id
        return new
    
    
    def copy_with(self, *, integration_id = ...):
        """
        Copies the role manager metadata with the given fields.
        
        Parameters
        ----------
        integration_id : `int`, ``Integration``, Optional (Keyword only)
            The manager integration's identifier.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # integration_id
        if integration_id is ...:
            integration_id = self.integration_id
        else:
            integration_id = validate_integration_id(integration_id)
        
        new = object.__new__(type(self))
        new.integration_id = integration_id
        return new
    
    
    @property
    @copy_docs(RoleManagerMetadataBase.manager_id)
    def manager_id(self):
        return self.integration_id
    
    
    @property
    @copy_docs(RoleManagerMetadataBase.manager)
    def manager(self):
        return create_partial_integration_from_id(self.integration_id)



__all__ = ()

from .constants import MAX_UINT_16, MAX_UINT_32, RTP_PACKET_TYPE_VOICE


def create_rtp_header_lite(sequence_number, timestamp, source):
    """
    Creates an rtp header data litely. Only the minimal fields.
    
    Parameters
    ----------
    sequence_number : `int`
        Sequence number.
    
    timestamp : `int`
        Timestamp.
    
    source : `int`
        Packet source.
    
    Returns
    -------
    data : `bytes`
    """
    return b''.join([
        b'\x80', # version = 2, nothing else
        RTP_PACKET_TYPE_VOICE.to_bytes(1, 'big'),
        (sequence_number & MAX_UINT_16).to_bytes(2, 'big'),
        (timestamp & MAX_UINT_32).to_bytes(4, 'big'),
        source.to_bytes(4, 'big'),
    ])


def create_rtp_data_lite(header, payload, padding):
    """
    Creates an rtp data litely.
    
    Parameters
    ----------
    header : `bytes`
        Header data.
    
    payload : `bytes-like`
        Payload
    
    padding : `None | bytes-like`
        Padding if any.
    
    Returns
    -------
    data : `bytes`
    """
    if (padding is None) or (not padding):
        return header + payload
    
    return b''.join([header, payload, padding])



from .constants import *
from .fields import *
from .welcome_screen_channel import *


__all__ = (
    *constants.__all__,
    *fields.__all__,
    *welcome_screen_channel.__all__,
)

# Apply deprecations ...

from ....utils.module_deprecation import deprecated_import

WelcomeChannel = deprecated_import(WelcomeScreenChannel, 'WelcomeChannel')



__all__ = ()

from warnings import warn

from scarletio import RichAttributeErrorBaseType

from .preinstanced import GuildFeature


class GuildBoostPerks(RichAttributeErrorBaseType):
    """
    Represents the boost perk that a user gets for each boost level.
    
    Attributes
    ----------
    attachment_size_limit : `int`
        The maximal size in bytes of each attachment that can be uploaded to the guild's channels.
    
    bitrate_limit : `int`
        The maximal bit rate usable in voice channel.
    
    concurrent_activities : `int`
        The maximal amount of concurrent embedded activities. Set as `0` if unlimited.
    
    emoji_limit : `int`
        The maximal amount of emojis.
    
    features : `None`, `tuple` of ``GuildFeature``
        The guild features acquired with the boost level.
    
    level : `int`
        The boost perks' level.
    
    screen_share_frame_rate : `int`
        The maximal frame rate usable when screen sharing.
    
    screen_share_resolution : `str`
        The maximal screen resolution's name when screen sharing.
    
    soundboard_sound_limit : `int`
        The maximal amount of sound board sounds.
    
    stage_channel_viewer_limit : `int`
        The amount of viewers the stage channels can have in the guild.
    
    sticker_limit : `int`
        The maximal amount of stickers the guild can have.
    """
    __slots__ = (
        'attachment_size_limit', 'bitrate_limit', 'concurrent_activities', 'emoji_limit', 'features', 'level',
        'screen_share_frame_rate', 'screen_share_resolution', 'soundboard_sound_limit', 'stage_channel_viewer_limit',
        'sticker_limit'
    )
    
    def __new__(
        cls,
        attachment_size_limit,
        bitrate_limit,
        concurrent_activities,
        emoji_limit,
        features,
        level,
        screen_share_frame_rate,
        screen_share_resolution,
        soundboard_sound_limit,
        stage_channel_viewer_limit,
        sticker_limit,
    ):
        """
        Creates a new guild boost perks instance.
        
        Parameters
        ----------
        attachment_size_limit : `int`
            The maximal size in bytes of each attachment that can be uploaded to the guild's channels.
        
        bitrate_limit : `int`
            The maximal bit rate usable in voice channel.
        
        concurrent_activities : `int`
            The maximal amount of concurrent embedded activities. Set as `0` if unlimited.
        
        emoji_limit : `int`
            The maximal amount of emojis.
        
        features : `None`, `tuple<GuildFeature>`
            The guild features acquired with the boost level.
        
        level : `int`
            The boost perks' level.
        
        screen_share_frame_rate : `int`
            The maximal frame rate usable when screen sharing.
        
        screen_share_resolution : `str`
            The maximal screen resolution's name when screen sharing.
        
        soundboard_sound_limit : `int`
            The maximal amount of sound board sounds.
        
        stage_channel_viewer_limit : `int`
            The amount of viewers the stage channels can have in the guild.
        
        sticker_limit : `int`
            The maximal amount of stickers the guild can have.
        """
        self = object.__new__(cls)
        self.attachment_size_limit = attachment_size_limit
        self.bitrate_limit = bitrate_limit
        self.concurrent_activities = concurrent_activities
        self.emoji_limit = emoji_limit
        self.features = features
        self.level = level
        self.screen_share_frame_rate = screen_share_frame_rate
        self.screen_share_resolution = screen_share_resolution
        self.soundboard_sound_limit = soundboard_sound_limit
        self.stage_channel_viewer_limit = stage_channel_viewer_limit
        self.sticker_limit = sticker_limit
        return self
    
    
    def __repr__(self):
        """Returns the guild premium perk's representation."""
        return f'{type(self).__name__} level = {self.level!r}>'
    
    
    def iter_features(self):
        """
        Iterates over the features of the guild premium perk.
        
        This method is an iterable generator.
        
        Yields
        ------
        feature : ``GuildFeature``
        """
        features = self.features
        if (features is not None):
            yield from features
    
    
    @property
    def upload_limit(self):
        """
        Deprecated and will be removed in 2025 November. Please use `.attachment_size_limit` instead.
        """
        warn(
            (
                f'`{type(self).__name__}.upload_limit` is deprecated and will be removed in 2025 November. '
                f'Please use `.attachment_size_limit` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        return self.boost_perks.attachment_size_limit
    
    
    @property
    def tier(self):
        """
        Deprecated and will be removed in 2025 November. Please use `.level` instead.
        """
        warn(
            (
                f'`{type(self).__name__}.tier` is deprecated and will be removed in 2025 November. '
                f'Please use `.level` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        return self.premium_perks.level


LEVEL_0 = GuildBoostPerks(
    8388608,
    96000,
    2,
    50,
    None,
    0,
    30,
    '720p',
    8,
    50,
    5,
)


LEVEL_1 = GuildBoostPerks(
    26214400,
    128000,
    3,
    100,
    (
        GuildFeature.icon_animated,
        GuildFeature.invite_splash,
    ),
    1,
    60,
    '720p',
    24,
    50,
    15,
)


LEVEL_2 = GuildBoostPerks(
    52428800,
    256000,
    5,
    150,
    (
        GuildFeature.banner,
        GuildFeature.icon_animated,
        GuildFeature.invite_splash,
        GuildFeature.role_icons,
    ),
    2,
    60,
    '1080p',
    36,
    150,
    30,
)


LEVEL_3 = GuildBoostPerks(
    104857600,
    384000,
    0,
    250,
    (
        GuildFeature.banner,
        GuildFeature.banner_animated,
        GuildFeature.icon_animated,
        GuildFeature.invite_splash,
        GuildFeature.invite_vanity_url,
        GuildFeature.role_icons,
    ),
    3,
    60,
    '1080p',
    48,
    300,
    60,
)


LEVELS = {
    0: LEVEL_0,
    1: LEVEL_1,
    2: LEVEL_2,
    3: LEVEL_3,
}

LEVEL_MAX = LEVEL_3



__all__ = ('AuditLogEntryType', 'AuditLogEntryTargetType')

from warnings import warn

from scarletio import class_property, copy_docs, export

from ...bases import Preinstance as P, PreinstancedBase

from ..audit_log_entry_change_conversions.application_command import (
    APPLICATION_COMMAND_CONVERSIONS as CHANGE_APPLICATION_COMMAND_CONVERSIONS
)
from ..audit_log_entry_change_conversions.auto_moderation_rule import (
    AUTO_MODERATION_RULE_CONVERSIONS as CHANGE_AUTO_MODERATION_RULE_CONVERSIONS
)
from ..audit_log_entry_change_conversions.channel import CHANNEL_CONVERSIONS as CHANGE_CHANNEL_CONVERSIONS
from ..audit_log_entry_change_conversions.channel_permission_overwrite import (
    CHANNEL_PERMISSION_OVERWRITE_CONVERSIONS as CHANGE_CHANNEL_PERMISSION_OVERWRITE_CONVERSIONS
)
from ..audit_log_entry_change_conversions.emoji import EMOJI_CONVERSIONS as CHANGE_EMOJI_CONVERSIONS
from ..audit_log_entry_change_conversions.guild import GUILD_CONVERSIONS as CHANGE_GUILD_CONVERSIONS
from ..audit_log_entry_change_conversions.integration import INTEGRATION_CONVERSIONS as CHANGE_INTEGRATION_CONVERSIONS
from ..audit_log_entry_change_conversions.invite import INVITE_CONVERSIONS as CHANGE_INVITE_CONVERSIONS
from ..audit_log_entry_change_conversions.onboarding_prompt import (
    ONBOARDING_PROMPT_CONVERSIONS as CHANGE_ONBOARDING_PROMPT_CONVERSIONS
)
from ..audit_log_entry_change_conversions.onboarding_screen import (
    ONBOARDING_SCREEN_CONVERSIONS as CHANGE_ONBOARDING_SCREEN_CONVERSIONS
)
from ..audit_log_entry_change_conversions.role import ROLE_CONVERSIONS as CHANGE_ROLE_CONVERSIONS
from ..audit_log_entry_change_conversions.scheduled_event import (
    SCHEDULED_EVENT_CONVERSIONS as CHANGE_SCHEDULED_EVENT_CONVERSIONS
)
from ..audit_log_entry_change_conversions.scheduled_event_occasion_overwrite import (
    SCHEDULED_EVENT_OCCASION_OVERWRITE_CONVERSIONS as CHANGE_SCHEDULED_EVENT_OCCASION_OVERWRITE_CONVERSIONS
)
from ..audit_log_entry_change_conversions.soundboard_sound import (
    SOUNDBOARD_SOUND_CONVERSIONS as CHANGE_SOUNDBOARD_SOUND_CONVERSIONS
)
from ..audit_log_entry_change_conversions.stage import STAGE_CONVERSIONS as CHANGE_STAGE_CONVERSIONS
from ..audit_log_entry_change_conversions.sticker import STICKER_CONVERSIONS as CHANGE_STICKER_CONVERSIONS
from ..audit_log_entry_change_conversions.user import USER_CONVERSIONS as CHANGE_USER_CONVERSIONS
from ..audit_log_entry_change_conversions.webhook import WEBHOOK_CONVERSIONS as CHANGE_WEBHOOK_CONVERSIONS
from ..audit_log_entry_detail_conversions.application_command import (
    APPLICATION_COMMAND_CONVERSIONS as DETAIL_APPLICATION_COMMAND_CONVERSIONS
)
from ..audit_log_entry_detail_conversions.auto_moderation_action_execution import (
    AUTO_MODERATION_ACTION_EXECUTION_CONVERSIONS as DETAIL_AUTO_MODERATION_ACTION_EXECUTION_CONVERSIONS
)
from ..audit_log_entry_detail_conversions.channel import CHANNEL_CONVERSIONS as DETAIL_CHANNEL_CONVERSIONS
from ..audit_log_entry_detail_conversions.channel_permission_overwrite import (
    CHANNEL_PERMISSION_OVERWRITE_CONVERSIONS as DETAIL_CHANNEL_PERMISSION_OVERWRITE_CONVERSIONS
)
from ..audit_log_entry_detail_conversions.guild import GUILD_CONVERSIONS as DETAIL_GUILD_CONVERSIONS
from ..audit_log_entry_detail_conversions.message import MESSAGE_CONVERSIONS as DETAIL_MESSAGE_CONVERSIONS
from ..audit_log_entry_detail_conversions.scheduled_event_occasion_overwrite import (
    SCHEDULED_EVENT_OCCASION_OVERWRITE_CONVERSIONS as DETAIL_SCHEDULED_EVENT_OCCASION_OVERWRITE_CONVERSIONS
)
from ..audit_log_entry_detail_conversions.stage import STAGE_CONVERSIONS as DETAIL_STAGE_CONVERSIONS
from ..audit_log_entry_detail_conversions.user import USER_CONVERSIONS as DETAIL_USER_CONVERSIONS

from .target_converters import (
    target_converter_application_command, target_converter_auto_moderation_rule, target_converter_channel,
    target_converter_emoji, target_converter_guild, target_converter_integration, target_converter_invite,
    target_converter_role, target_converter_scheduled_event, target_converter_soundboard_sound, target_converter_stage,
    target_converter_sticker, target_converter_user, target_converter_webhook
)


class AuditLogEntryTargetType(PreinstancedBase, value_type = int):
    """
    Represents the target type of an ``AuditLogEntry``.
    
    Attributes
    ----------
    change_conversions : `None | AuditLogEntryDetailConversionGroup`
        Change conversions.
    
    detail_conversions : `None | AuditLogEntryDetailConversionGroup`
        Detail conversions.
    
    name : `str`
        The name of audit log target type.
    
    target_converter : `None | FunctionType | MethodType`
        Audit log target converter.
    
    value : `int`
        The identifier value of the audit log target type. Only used for hashing.
    
    Type Attributes
    ---------------
    Every predefined audit log target type can be accessed as type attribute as well:
    
    +---------------------------------------+---------------------------------------+-------+
    | Type attribute name                   | name                                  | value |
    +=======================================+=======================================+=======+
    | none                                  | none                                  | 0     |
    +---------------------------------------+---------------------------------------+-------+
    | guild                                 | guild                                 | 1     |
    +---------------------------------------+---------------------------------------+-------+
    | channel                               | channel                               | 2     |
    +---------------------------------------+---------------------------------------+-------+
    | channel_permission_overwrite          | channel permission overwrite          | 3     |
    +---------------------------------------+---------------------------------------+-------+
    | user                                  | user                                  | 4     |
    +---------------------------------------+---------------------------------------+-------+
    | role                                  | role                                  | 5     |
    +---------------------------------------+---------------------------------------+-------+
    | invite                                | invite                                | 6     |
    +---------------------------------------+---------------------------------------+-------+
    | webhook                               | webhook                               | 7     |
    +---------------------------------------+---------------------------------------+-------+
    | emoji                                 | emoji                                 | 8     |
    +---------------------------------------+---------------------------------------+-------+
    | integration                           | integration                           | 9     |
    +---------------------------------------+---------------------------------------+-------+
    | stage                                 | stage                                 | 10    |
    +---------------------------------------+---------------------------------------+-------+
    | scheduled_event                       | scheduled event                       | 11    |
    +---------------------------------------+---------------------------------------+-------+
    | sticker                               | sticker                               | 12    |
    +---------------------------------------+---------------------------------------+-------+
    | application_command                   | application command                   | 13    |
    +---------------------------------------+---------------------------------------+-------+
    | auto_moderation_rule                  | auto moderation rule                  | 14    |
    +---------------------------------------+---------------------------------------+-------+
    | auto_moderation_action_execution      | auto moderation action execution      | 15    |
    +---------------------------------------+---------------------------------------+-------+
    | onboarding_screen                     | onboarding screen                     | 16    |
    +---------------------------------------+---------------------------------------+-------+
    | onboarding_prompt                     | onboarding prompt                     | 17    |
    +---------------------------------------+---------------------------------------+-------+
    | message                               | message                               | 18    |
    +---------------------------------------+---------------------------------------+-------+
    | soundboard_sound                      | soundboard sound                      | 19    |
    +---------------------------------------+---------------------------------------+-------+
    | scheduled_event_occasion_overwrite    | scheduled event occasion overwrite    | 20    |
    +---------------------------------------+---------------------------------------+-------+
    """
    __slots__ = ('change_conversions', 'detail_conversions', 'target_converter')
    
    def __new__(cls, value, name = None, target_converter = None, change_conversions = None, detail_conversions = None):
        """
        Creates a new audit log target type.
        
        Parameters
        ----------
        value : `int`
            The identifier value of the audit log target type.
        
        name : `None | str` = `None`, Optional
            The name of the audit log target type.
        
        target_converter : `None | FunctionType | MethodType` = `None`, Optional
            Audit log target converter.
        
        change_conversions :  `None | AuditLogEntryChangeConversionGroup` = `None`, Optional
            Change conversions.
        
        detail_conversions : `None | AuditLogEntryDetailConversionGroup` = `None`, Optional
            Detail conversions.
        """
        self = PreinstancedBase.__new__(cls, value, name)
        self.target_converter = target_converter
        self.change_conversions = change_conversions
        self.detail_conversions = detail_conversions
        return self
    

    # predefined
    none = P(
        0,
        'none',
        None,
        None,
        None,
    )
    guild = P(
        1,
        'guild',
        target_converter_guild,
        CHANGE_GUILD_CONVERSIONS,
        DETAIL_GUILD_CONVERSIONS,
    )
    channel = P(
        2,
        'channel',
        target_converter_channel,
        CHANGE_CHANNEL_CONVERSIONS,
        DETAIL_CHANNEL_CONVERSIONS,
    )
    channel_permission_overwrite = P(
        3,
        'channel permission overwrite',
        target_converter_channel,
        CHANGE_CHANNEL_PERMISSION_OVERWRITE_CONVERSIONS,
        DETAIL_CHANNEL_PERMISSION_OVERWRITE_CONVERSIONS,
    )
    user = P(
        4,
        'user',
        target_converter_user,
        CHANGE_USER_CONVERSIONS,
        DETAIL_USER_CONVERSIONS,
    )
    role = P(
        5,
        'role',
        target_converter_role,
        CHANGE_ROLE_CONVERSIONS,
        None,
    )
    invite = P(
        6,
        'invite',
        target_converter_invite,
        CHANGE_INVITE_CONVERSIONS,
        None,
    )
    webhook = P(
        7,
        'webhook',
        target_converter_webhook,
        CHANGE_WEBHOOK_CONVERSIONS,
        None,
    )
    emoji = P(
        8,
        'emoji',
        target_converter_emoji,
        CHANGE_EMOJI_CONVERSIONS,
        None,
    )
    integration = P(
        9,
        'integration',
        target_converter_integration,
        CHANGE_INTEGRATION_CONVERSIONS,
        None,
    )
    stage = P(
        10,
        'stage',
        target_converter_stage,
        CHANGE_STAGE_CONVERSIONS,
        DETAIL_STAGE_CONVERSIONS,
    )
    scheduled_event = P(
        11,
        'scheduled event',
        target_converter_scheduled_event,
        CHANGE_SCHEDULED_EVENT_CONVERSIONS,
        None,
    )
    sticker = P(
        12,
        'sticker',
        target_converter_sticker,
        CHANGE_STICKER_CONVERSIONS,
        None,
    )
    application_command = P(
        13,
        'application command',
        target_converter_application_command,
        CHANGE_APPLICATION_COMMAND_CONVERSIONS,
        DETAIL_APPLICATION_COMMAND_CONVERSIONS,
    )
    auto_moderation_rule = P(
        14,
        'auto moderation rule',
        target_converter_auto_moderation_rule,
        CHANGE_AUTO_MODERATION_RULE_CONVERSIONS,
        None,
    )
    auto_moderation_action_execution = P(
        15,
        'auto moderation action execution',
        None,
        CHANGE_USER_CONVERSIONS,
        DETAIL_AUTO_MODERATION_ACTION_EXECUTION_CONVERSIONS,
    )
    onboarding_screen = P(
        16,
        'onboarding screen',
        None,
        CHANGE_ONBOARDING_SCREEN_CONVERSIONS,
        None,
    )
    onboarding_prompt = P(
        17,
        'onboarding prompt',
        None,
        CHANGE_ONBOARDING_PROMPT_CONVERSIONS,
        None,
    )
    message = P(
        18,
        'message',
        None,
        None,
        DETAIL_MESSAGE_CONVERSIONS,
    )
    soundboard_sound = P(
        19,
        'soundboard sound',
        target_converter_soundboard_sound,
        CHANGE_SOUNDBOARD_SOUND_CONVERSIONS,
        None,
    )
    scheduled_event_occasion_overwrite = P(
        20,
        'scheduled event occasion overwrite',
        target_converter_scheduled_event,
        CHANGE_SCHEDULED_EVENT_OCCASION_OVERWRITE_CONVERSIONS,
        DETAIL_SCHEDULED_EVENT_OCCASION_OVERWRITE_CONVERSIONS,
    )


@export
class AuditLogEntryType(PreinstancedBase, value_type = int):
    """
    Represents the event type of an ``AuditLogEntry``.
    
    Attributes
    ----------
    name : `str`
        The name of audit log event.
    
    target_type : ``AuditLogEntryTargetType``
        The audit log's target's type.
    
    value : `int`
        The Discord side identifier value of the audit log event.
    
    Type Attributes
    ---------------
    Every predefined audit log event can be accessed as type attribute as well:
    
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | Type attribute name                               | Name                                              | Value | Target type                           |
    +===================================================+===================================================+=======+=======================================+
    | none                                              | none                                              |  0    | none                                  |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | guild_update                                      | guild update                                      |  1    | guild                                 |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | channel_create                                    | channel create                                    | 10    | channel                               |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | channel_update                                    | channel update                                    | 11    | channel                               |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | channel_delete                                    | channel delete                                    | 12    | channel                               |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | channel_permission_overwrite_create               | channel permission overwrite create               | 13    | channel_permission_overwrite          |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | channel_permission_overwrite_update               | channel permission overwrite update               | 14    | channel_permission_overwrite          |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | channel_permission_overwrite_delete               | channel permission overwrite delete               | 15    | channel_permission_overwrite          |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | user_kick                                         | user kick                                         | 20    | user                                  |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | user_prune                                        | user prune                                        | 21    | guild                                 |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | user_ban_add                                      | user ban add                                      | 22    | user                                  |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | user_ban_remove                                   | user ban remove                                   | 23    | user                                  |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | user_update                                       | user update                                       | 24    | user                                  |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | user_role_update                                  | user role update                                  | 25    | user                                  |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | user_move                                         | user move                                         | 26    | user                                  |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | user_disconnect                                   | user disconnect                                   | 27    | user                                  |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | bot_add                                           | bot add                                           | 28    | user                                  |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | role_create                                       | role create                                       | 30    | role                                  |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | role_update                                       | role update                                       | 31    | role                                  |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | role_delete                                       | role delete                                       | 32    | role                                  |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | invite_create                                     | invite create                                     | 40    | invite                                |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | invite_update                                     | invite update                                     | 41    | invite                                |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | invite_delete                                     | invite delete                                     | 42    | invite                                |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | webhook_create                                    | webhook create                                    | 50    | webhook                               |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | webhook_update                                    | webhook update                                    | 51    | webhook                               |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | webhook_delete                                    | webhook delete                                    | 52    | webhook                               |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | emoji_create                                      | emoji create                                      | 60    | emoji                                 |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | emoji_update                                      | emoji update                                      | 61    | emoji                                 |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | emoji_delete                                      | emoji delete                                      | 62    | emoji                                 |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | message_delete                                    | message delete                                    | 72    | message                               |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | message_bulk_delete                               | message bulk delete                               | 73    | message                               |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | message_pin                                       | message pin                                       | 74    | message                               |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | message_unpin                                     | message unpin                                     | 75    | message                               |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | integration_create                                | integration create                                | 80    | integration                           |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | integration_update                                | integration update                                | 81    | integration                           |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | integration_delete                                | integration delete                                | 82    | integration                           |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | stage_create                                      | stage create                                      | 83    | stage                                 |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | stage_update                                      | stage update                                      | 84    | stage                                 |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | stage_delete                                      | stage delete                                      | 85    | stage                                 |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | sticker_create                                    | sticker create                                    | 90    | sticker                               |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | sticker_update                                    | sticker update                                    | 91    | sticker                               |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | sticker_delete                                    | sticker delete                                    | 92    | sticker                               |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | scheduled_event_create                            | scheduled event create                            | 100   | scheduled_event                       |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | scheduled_event_update                            | scheduled event update                            | 101   | scheduled_event                       |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | scheduled_event_delete                            | scheduled event delete                            | 102   | scheduled_event                       |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | thread_create                                     | thread create                                     | 110   | channel                               |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | thread_update                                     | thread update                                     | 111   | channel                               |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | thread_delete                                     | thread delete                                     | 112   | channel                               |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | application_command_permission_update             | application command permission update             | 121   | application_command                   |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | soundboard_sound_create                           | soundboard sound create                           | 130   | soundboard_sound                      |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | soundboard_sound_update                           | soundboard sound update                           | 131   | soundboard_sound                      |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | soundboard_sound_delete                           | soundboard sound delete                           | 132   | soundboard_sound                      |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | auto_moderation_rule_create                       | auto moderation rule create                       | 140   | auto_moderation_rule                  |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | auto_moderation_rule_update                       | auto moderation rule update                       | 141   | auto_moderation_rule                  |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | auto_moderation_rule_delete                       | auto moderation rule delete                       | 142   | auto_moderation_rule                  |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | auto_moderation_block_message                     | auto moderation block message                     | 143   | auto_moderation_action_execution      |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | auto_moderation_alert_message                     | auto moderation alert message                     | 144   | auto_moderation_action_execution      |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | auto_moderation_user_timeout                      | auto moderation user timeout                      | 145   | auto_moderation_action_execution      |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | auto_moderation_quarantined_user                  | auto moderation quarantined user                  | 146   | auto_moderation_action_execution      |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | creator_monetization_request_created              | creator monetization request created              | 150   | none                                  |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | creator_monetization_terms_accepted               | creator monetization terms accepted               | 151   | none                                  |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | role_prompt_create                                | role_prompt_create                                | 160   | none                                  |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | role_prompt_update                                | role_prompt_update                                | 161   | none                                  |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | role_prompt_delete                                | role_prompt_delete                                | 162   | none                                  |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | onboarding_prompt_create                          | onboarding prompt create                          | 163   | onboarding_prompt                     |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | onboarding_prompt_update                          | onboarding prompt update                          | 164   | onboarding_prompt                     |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | onboarding_prompt_delete                          | onboarding prompt delete                          | 165   | onboarding_prompt                     |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | onboarding_screen_create                          | onboarding screen create                          | 166   | onboarding_screen                     |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | onboarding_screen_update                          | onboarding screen update                          | 167   | onboarding_screen                     |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | home_feature_item                                 | home feature item                                 | 171   | none                                  |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | home_remove_item                                  | home remove item                                  | 172   | none                                  |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | home_screen_create                                | home screen create                                | 190   | none                                  |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | home_screen_update                                | home screen update                                | 191   | none                                  |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | channel_status_update                             | channel status update                             | 192   | channel                               |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | channel_status_delete                             | channel status delete                             | 193   | channel                               |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | clyde_ai_update                                   | clyde ai update                                   | 194   | none                                  |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | scheduled_event_occasion_overwrite_create         | scheduled event occasion overwrite create         | 200   | scheduled_event_occasion_overwrite    |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | scheduled_event_occasion_overwrite_update         | scheduled event occasion overwrite update         | 201   | scheduled_event_occasion_overwrite    |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | scheduled_event_occasion_overwrite_delete         | scheduled event occasion overwrite delete         | 202   | scheduled_event_occasion_overwrite    |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | user_verification_update                          | user verification update                          | 210   | none                                  |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | user_self_update                                  | user self update                                  | 211   | user                                  |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | permission_migration_pin_completed                | permission migration pin completed                | 212   | user                                  |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    | permission_migration_slowmode_bypass_completed    | permission migration slowmode bypass completed    | 213   | none                                  |
    +---------------------------------------------------+---------------------------------------------------+-------+---------------------------------------+
    """
    __slots__ = ('target_type',)
    
    def __new__(cls, value, name = None, target_type = None):
        """
        Creates a new audit log event.
        
        Parameters
        ----------
        value : `int`
            The identifier value of the audit log event.
        
        name : `None | str` = `None`, Optional
            The audit log event's name.
        
        target_type : `None | AuditLogEntryTargetType` = `None`, Optional
            The audit log's target type.
        """
        if target_type is None:
            target_type = AuditLogEntryTargetType.none
        
        self = PreinstancedBase.__new__(cls, value, name)
        self.target_type = target_type
        return self
    
    
    @classmethod
    def _from_value(cls, value):
        """
        Creates a new audit log event object from the given value.
        
        Parameters
        ----------
        value : ``.VALUE_TYPE``
            The value what has no representation yet.
        
        Returns
        -------
        self : ``AuditLogEntryType``
            The created audit log event.
        """
        self = super(AuditLogEntryType, cls)._from_value(value)
        self.target_type = AuditLogEntryTargetType.none
        return self
    
    
    @copy_docs(PreinstancedBase._put_repr_parts_into)
    def _put_repr_parts_into(self, repr_parts):
        repr_parts.append(', target_type = ')
        repr_parts.append(self.target_type.name)
    
    
    # predefined
    none = P(0, 'none', AuditLogEntryTargetType.none)
    
    guild_update = P(1, 'guild update', AuditLogEntryTargetType.guild)
    
    channel_create = P(10, 'channel create', AuditLogEntryTargetType.channel)
    channel_update = P(11, 'channel update', AuditLogEntryTargetType.channel)
    channel_delete = P(12, 'channel delete', AuditLogEntryTargetType.channel)
    
    channel_permission_overwrite_create = P(
        13,
        'channel permission overwrite create',
        AuditLogEntryTargetType.channel_permission_overwrite,
    )
    channel_permission_overwrite_update = P(
        14,
        'channel permission overwrite update',
        AuditLogEntryTargetType.channel_permission_overwrite,
    )
    channel_permission_overwrite_delete = P(
        15,
        'channel permission overwrite delete',
        AuditLogEntryTargetType.channel_permission_overwrite,
    )
    
    user_kick = P(20, 'user kick', AuditLogEntryTargetType.user)
    user_prune = P(21, 'user prune', AuditLogEntryTargetType.guild)
    user_ban_add = P(22, 'user ban add', AuditLogEntryTargetType.user)
    user_ban_remove = P(23, 'user ban remove', AuditLogEntryTargetType.user)
    user_update = P(24, 'user update', AuditLogEntryTargetType.user)
    user_role_update = P(25, 'user role update', AuditLogEntryTargetType.user)
    user_move = P(26, 'user move', AuditLogEntryTargetType.user)
    user_disconnect = P(27, 'user disconnect', AuditLogEntryTargetType.user)
    bot_add = P(28, 'user role update', AuditLogEntryTargetType.user)
    
    role_create = P(30, 'role create', AuditLogEntryTargetType.role)
    role_update = P(31, 'role update', AuditLogEntryTargetType.role)
    role_delete = P(32, 'role delete', AuditLogEntryTargetType.role)
    
    invite_create = P(40, 'invite create', AuditLogEntryTargetType.invite)
    invite_update = P(41, 'invite update', AuditLogEntryTargetType.invite)
    invite_delete = P(42, 'invite delete', AuditLogEntryTargetType.invite)
    
    webhook_create = P(50, 'webhook create', AuditLogEntryTargetType.webhook)
    webhook_update = P(51, 'webhook update', AuditLogEntryTargetType.webhook)
    webhook_delete = P(52, 'webhook delete', AuditLogEntryTargetType.webhook)
    
    emoji_create = P(60, 'emoji create', AuditLogEntryTargetType.emoji)
    emoji_update = P(61, 'emoji update', AuditLogEntryTargetType.emoji)
    emoji_delete = P(62, 'emoji delete', AuditLogEntryTargetType.emoji)
    
    message_delete = P(72, 'message delete', AuditLogEntryTargetType.message)
    message_bulk_delete = P(73, 'message bulk delete', AuditLogEntryTargetType.message)
    message_pin = P(74, 'message pin', AuditLogEntryTargetType.message)
    message_unpin = P(75, 'message unpin', AuditLogEntryTargetType.message)
    
    integration_create = P(80, 'integration create', AuditLogEntryTargetType.integration)
    integration_update = P(81, 'integration update', AuditLogEntryTargetType.integration)
    integration_delete = P(82, 'integration delete', AuditLogEntryTargetType.integration)
    
    stage_create = P(83, 'stage create', AuditLogEntryTargetType.stage)
    stage_update = P(84, 'stage update', AuditLogEntryTargetType.stage)
    stage_delete = P(85, 'stage delete', AuditLogEntryTargetType.stage)
    
    sticker_create = P(90, 'sticker create', AuditLogEntryTargetType.sticker)
    sticker_update = P(91, 'sticker update', AuditLogEntryTargetType.sticker)
    sticker_delete = P(92, 'sticker delete', AuditLogEntryTargetType.sticker)
    
    scheduled_event_create = P(100, 'scheduled event create', AuditLogEntryTargetType.scheduled_event)
    scheduled_event_update = P(101, 'scheduled event update', AuditLogEntryTargetType.scheduled_event)
    scheduled_event_delete = P(102, 'scheduled event delete', AuditLogEntryTargetType.scheduled_event)
    
    thread_create = P(110, 'thread create', AuditLogEntryTargetType.channel)
    thread_update = P(111, 'thread update', AuditLogEntryTargetType.channel)
    thread_delete = P(112, 'thread delete', AuditLogEntryTargetType.channel)
    
    application_command_permission_update = P(
        121,
        'application command permission update',
        AuditLogEntryTargetType.application_command,
    )
    
    soundboard_sound_create =  P(130,'soundboard sound create', AuditLogEntryTargetType.soundboard_sound)
    soundboard_sound_update =  P(131, 'soundboard sound update', AuditLogEntryTargetType.soundboard_sound)
    soundboard_sound_delete = P(132, 'soundboard sound delete', AuditLogEntryTargetType.soundboard_sound)
    
    auto_moderation_rule_create = P(140, 'auto moderation rule create', AuditLogEntryTargetType.auto_moderation_rule)
    auto_moderation_rule_update = P(141, 'auto moderation rule update', AuditLogEntryTargetType.auto_moderation_rule)
    auto_moderation_rule_delete = P(142, 'auto moderation rule delete', AuditLogEntryTargetType.auto_moderation_rule)
    auto_moderation_block_message = P(
        143, 'auto moderation block message', AuditLogEntryTargetType.auto_moderation_action_execution
    )
    auto_moderation_alert_message = P(
        144, 'auto moderation alert message', AuditLogEntryTargetType.auto_moderation_action_execution
    )
    auto_moderation_user_timeout = P(
        145, 'auto moderation user timeout', AuditLogEntryTargetType.auto_moderation_action_execution
    )
    auto_moderation_quarantined_user = P(
        146, 'auto moderation quarantined user', AuditLogEntryTargetType.auto_moderation_action_execution
    )
    
    creator_monetization_request_created = P(150, 'creator monetization request created', AuditLogEntryTargetType.none)
    creator_monetization_terms_accepted = P(151, 'creator monetization terms accepted', AuditLogEntryTargetType.none)
    
    # These 3 are deprecated
    role_prompt_create =  P(160,'role prompt create', AuditLogEntryTargetType.none)
    role_prompt_update =  P(161, 'role prompt update', AuditLogEntryTargetType.none)
    role_prompt_delete = P(162, 'role prompt delete', AuditLogEntryTargetType.none)
    
    onboarding_prompt_create =  P(163,'onboarding prompt create', AuditLogEntryTargetType.onboarding_prompt)
    onboarding_prompt_update =  P(164, 'onboarding prompt update', AuditLogEntryTargetType.onboarding_prompt)
    onboarding_prompt_delete = P(165, 'onboarding prompt delete', AuditLogEntryTargetType.onboarding_prompt)
    
    onboarding_screen_create =  P(166,'onboarding screen create', AuditLogEntryTargetType.onboarding_screen)
    onboarding_screen_update =  P(167, 'onboarding screen update', AuditLogEntryTargetType.onboarding_screen)
    
    home_feature_item = P(171, 'home feature item', AuditLogEntryTargetType.none)
    home_remove_item = P(172, 'home remove item', AuditLogEntryTargetType.none)
    
    home_screen_create = P(190, 'home screen create', AuditLogEntryTargetType.none)
    home_screen_update = P(191, 'home screen update', AuditLogEntryTargetType.none)
    channel_status_update = P(192, 'channel status update', AuditLogEntryTargetType.channel)
    channel_status_delete = P(193, 'channel status delete', AuditLogEntryTargetType.channel)
    clyde_ai_update = P(194, 'clyde ai update', AuditLogEntryTargetType.none)
    
    scheduled_event_occasion_overwrite_create = P(
        200,
        'scheduled event occasion overwrite create',
        AuditLogEntryTargetType.scheduled_event_occasion_overwrite,
    )
    scheduled_event_occasion_overwrite_update = P(
        201,
        'scheduled event occasion overwrite update',
        AuditLogEntryTargetType.scheduled_event_occasion_overwrite,
    )
    scheduled_event_occasion_overwrite_delete = P(
        202,
        'scheduled event occasion overwrite delete',
        AuditLogEntryTargetType.scheduled_event_occasion_overwrite,
    )
    
    user_verification_update = P(210, 'user verification update', AuditLogEntryTargetType.none)
    user_self_update = P(211, 'user self update', AuditLogEntryTargetType.user)
    permission_migration_pin_completed = P(
        122,
        'permission migration pin completed',
        AuditLogEntryTargetType.none,
    )
    permission_migration_slowmode_bypass_completed = P(
        123,
        'permission migration slowmode bypass completed',
        AuditLogEntryTargetType.none,
    )
    
    
    @class_property
    def scheduled_event_exception_create(cls):
        """
        Deprecated and will be removed in 2016 January. Use `scheduled_event_occasion_overwrite_create.` instead.
        """
        warn(
            (
                f'`{cls.__name__}.scheduled_event_exception_create` is deprecated and will be removed 2016 January. '
                f'Use `.scheduled_event_occasion_overwrite_create` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        
        return cls.scheduled_event_occasion_overwrite_create
    
    
    @class_property
    def scheduled_event_exception_update(cls):
        """
        Deprecated and will be removed in 2016 January. Use `scheduled_event_occasion_overwrite_update.` instead.
        """
        warn(
            (
                f'`{cls.__name__}.scheduled_event_exception_update` is deprecated and will be removed 2016 January. '
                f'Use `.scheduled_event_occasion_overwrite_update` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        return cls.scheduled_event_occasion_overwrite_update
    
    
    @class_property
    def scheduled_event_exception_delete(cls):
        """
        Deprecated and will be removed in 2016 January. Use `scheduled_event_occasion_overwrite_delete.` instead.
        """
        warn(
            (
                f'`{cls.__name__}.scheduled_event_exception_delete` is deprecated and will be removed 2016 January. '
                f'Use `.scheduled_event_occasion_overwrite_delete` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        
        return cls.scheduled_event_occasion_overwrite_delete



__all__ = ('ActivityChange',)

from scarletio import RichAttributeErrorBaseType

from .fields import validate_added, validate_updated, validate_removed


class ActivityChange(RichAttributeErrorBaseType):
    """
    Represents a user's changed activities.
    
    Attributes
    ----------
    added : ``None | list<Activity>``
        The added activities to the respective user. Defaults to `None`.
    
    removed: ``None | list<Activity>``
        The removed activities from the respective user. Defaults to `None`.
    
    updated : `None`, `list` of ``ActivityUpdate``
        The updated activities of the respective user. Defaults to `None`.
    """
    __slots__ = ('added', 'updated', 'removed',)
    
    def __new__(cls, *, added = ..., removed = ..., updated = ...):
        """
        Creates a new activity change with the given parameters.
        
        Parameters
        ----------
        added : `None`, `iterable` of ``Activity``, Optional (Keyword only)
            The added activities to the user.
        
        removed: `None`, `iterable` of ``Activity``, Optional (Keyword only)
            The removed activities from the user.
        
        updated : `None`, `iterable` of ``ActivityUpdate``, Optional (Keyword only)
            The updated activities of the user.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # added
        if added is ...:
            added = None
        else:
            added = validate_added(added)
        
        # removed
        if removed is ...:
            removed = None
        else:
            removed = validate_removed(removed)
        
        # updated
        if updated is ...:
            updated = None
        else:
            updated = validate_updated(updated)
        
        # Construct
        self = object.__new__(cls)
        self.added = added
        self.removed = removed
        self.updated = updated
        return self
    
    
    @classmethod
    def from_fields(cls, added, updated, removed):
        """
        Creates a new activity change with the given parameters.
        
        Parameters
        ----------
        added : ``None | list<Activity>``
            The added activities to the user.
        updated : `None`, `list` of ``ActivityUpdate``
            The updated activities of the user.
        removed: ``None | list<Activity>``
            The removed activities from the user.
        """
        self = object.__new__(cls)
        self.added = added
        self.removed = removed
        self.updated = updated
        return self
    
    
    def __repr__(self):
        """Returns the representation of the activity change."""
        repr_parts = ['<',
            self.__class__.__name__,
        ]
        
        added = self.added
        if added is None:
            field_added = False
        else:
            repr_parts.append(' added = ')
            repr_parts.append(repr(added))
            field_added = True
        
        updated = self.updated
        if (updated is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' updated = ')
            repr_parts.append(repr(updated))
        
        removed = self.removed
        if (removed is not None):
            if field_added:
                repr_parts.append(',')
            
            repr_parts.append(' removed = ')
            repr_parts.append(repr(removed))
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns the activity change's hash value."""
        hash_value = 0
        
        added = self.added
        if (added is not None):
            hash_value ^= len(added)
            
            for activity in added:
                hash_value ^= hash(activity)
        
        removed = self.removed
        if (removed is not None):
            hash_value ^= len(removed) << 8
            
            for activity in removed:
                hash_value ^= hash(activity)
        
        updated = self.updated
        if (updated is not None):
            hash_value ^= len(updated) << 4
            
            for activity_update in updated:
                hash_value ^= hash(activity_update)
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two activity changes are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.added != other.added:
            return False
        
        if self.removed != other.removed:
            return False
        
        if self.updated != other.updated:
            return False
        
        return True
    
    
    def __len__(self):
        """Helper for unpacking if needed."""
        return 3
    
    
    def __iter__(self):
        """
        Unpacks the activity change.
        
        This method is a generator.
        """
        yield self.added
        yield self.updated
        yield self.removed
    
    
    def iter_added(self):
        """
        Iterates over the added activities.
        
        This method is an iterable generator.
        
        Yields
        ------
        added : ``Activity``
        """
        added = self.added
        if (added is not None):
            yield from added
    
    
    def iter_updated(self):
        """
        Iterates over the activity updates.
        
        This method is an iterable generator.
        
        Yields
        ------
        updated : ``ActivityUpdate``
        """
        updated = self.updated
        if (updated is not None):
            yield from updated
    
    
    def iter_removed(self):
        """
        Iterates over the removed activities.
        
        This method is an iterable generator.
        
        Yields
        ------
        removed : ``Activity``
        """
        removed = self.removed
        if (removed is not None):
            yield from removed
    
    
    def copy(self):
        """
        Copies the activity change.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        added = self.added
        if (added is not None):
            added = [activity.copy() for activity in added]
        
        removed = self.removed
        if (removed is not None):
            removed = [activity.copy() for activity in removed]
        
        updated = self.updated
        if (updated is not None):
            updated = [activity_update.copy() for activity_update in updated]
        
        new = object.__new__(type(self))
        new.added = added
        new.updated = updated
        new.removed = removed
        return new
    
    
    def copy_with(self, *, added = ..., removed = ..., updated = ...):
        """
        Copies the activity change with the given fields.
        
        Parameters
        ----------
        added : `None`, `iterable` of ``Activity``, Optional (Keyword only)
            The added activities to the user.
        
        removed: `None`, `iterable` of ``Activity``, Optional (Keyword only)
            The removed activities from the user.
        
        updated : `None`, `iterable` of ``ActivityUpdate``, Optional (Keyword only)
            The updated activities of the user.
        
        Returns
        -------
        new : `instance<type<self>>`
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # added
        if added is ...:
            added = self.added
            if (added is not None):
                added = [activity.copy() for activity in added]
        else:
            added = validate_added(added)
        
        # removed
        if removed is ...:
            removed = self.removed
            if (removed is not None):
                removed = [activity.copy() for activity in removed]
        else:
            removed = validate_removed(removed)
        
        # updated
        if updated is ...:
            updated = self.updated
            if (updated is not None):
                updated = [activity_update.copy() for activity_update in updated]
        else:
            updated = validate_updated(updated)
        
        # Construct
        new = object.__new__(type(self))
        new.added = added
        new.removed = removed
        new.updated = updated
        return new



__all__ = ('EmbedField',)

from scarletio import copy_docs

from ...utils import sanitize_mentions

from ..embed_field_base import EmbedFieldBase

from .fields import (
    parse_inline, parse_name, parse_value, put_inline, put_name, put_value, validate_inline,
    validate_name, validate_value
)


class EmbedField(EmbedFieldBase):
    """
    Represents an embed's author.
    
    Attributes
    ----------
    inline : `bool`
        Whether this field should display inline. Defaults to `False`.
    name : `None`, `str`
        The name of the field.
    value : `None`, `str`
        The value of the field.
    """
    __slots__ = ('value', 'name', 'inline')
    
    def __new__(cls, name = ..., value = ..., inline = ...):
        """
        Creates a new embed field with the given parameters.
        
        Parameters
        ----------
        name : `None`, `str`, Optional
            The name of the field.
        value : `None`, `str`, Optional
            The value of the field.
        inline : `bool`, Optional
            Whether this field should display inline.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # inline
        if inline is ...:
            inline = False
        else:
            inline = validate_inline(inline)
        
        # name
        if name is ...:
            name = None
        else:
            name = validate_name(name)
        
        # value
        if value is ...:
            value = None
        else:
            value = validate_value(value)
        
        self = object.__new__(cls)
        self.inline = inline
        self.name = name
        self.value = value
        return self
    
        
    @copy_docs(EmbedFieldBase.__len__)
    def __len__(self):
        length = 0
        
        # name
        name = self.name
        if (name is not None):
            length += len(name)
        
        # value
        value = self.value
        if (value is not None):
            length += len(value)
        
        return length
    
    
    @copy_docs(EmbedFieldBase.__bool__)
    def __bool__(self):
        # name
        if (self.name is not None):
            return True
        
        # value
        if (self.value is not None):
            return True
        
        return False
    
    
    @copy_docs(EmbedFieldBase._put_repr_parts_into)
    def _put_repr_parts_into(self, repr_parts):
        field_added = False
        
        name = self.name
        if name is not None:
            field_added = True
            repr_parts.append(' name = ')
            repr_parts.append(repr(self.name))
        
        value = self.value
        if value is not None:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' value = ')
            repr_parts.append(repr(value))
        
        inline = self.inline
        if inline:
            if field_added:
                repr_parts.append(',')
            
            repr_parts.append(' inline = ')
            repr_parts.append(repr(inline))
    
    
    @copy_docs(EmbedFieldBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # inline
        hash_value ^= hash(self.inline)
        
        # name
        name = self.name
        if (name is not None):
            hash_value ^= hash(name)
        
        # value
        value = self.value
        if (value is not None) and (name != value):
            hash_value ^= hash(value)
        
        return hash_value
    
    
    @copy_docs(EmbedFieldBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        # inline
        if self.inline != other.inline:
            return False
        
        # name
        if self.name != other.name:
            return False
        
        # value
        if self.value != other.value:
            return False
        
        return True
    
    
    @classmethod
    @copy_docs(EmbedFieldBase.from_data)
    def from_data(cls, data):
        self = object.__new__(cls)
        self.inline = parse_inline(data)
        self.name = parse_name(data)
        self.value = parse_value(data)
        return self
    
    
    @copy_docs(EmbedFieldBase.to_data)
    def to_data(self, *, defaults = False, include_internals = False):
        data = {}
        put_inline(self.inline, data, defaults)
        put_name(self.name, data, defaults)
        put_value(self.value, data, defaults)
        return data
    
    
    @copy_docs(EmbedFieldBase.clean_copy)
    def clean_copy(self, guild = None):
        new = object.__new__(type(self))
        new.inline = self.inline
        new.name = sanitize_mentions(self.name, guild)
        new.value = sanitize_mentions(self.value, guild)
        return new
    
    
    @copy_docs(EmbedFieldBase.copy)
    def copy(self):
        new = object.__new__(type(self))
        new.inline = self.inline
        new.name = self.name
        new.value = self.value
        return new
    
    
    def copy_with(self, *, inline = ..., name = ..., value = ...):
        """
        Copies the embed author with the given parameters.
        
        Parameters
        ----------
        inline : `bool`, Optional (Keyword only)
            Whether this field should display inline.
        name : `None`, `str`, Optional (Keyword only)
            The name of the field.
        value : `None`, `str`, Optional (Keyword only)
            The value of the field.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # inline
        if inline is ...:
            inline = self.inline
        else:
            inline = validate_inline(inline)
        
        # name
        if name is ...:
            name = self.name
        else:
            name = validate_name(name)
        
        # value
        if value is ...:
            value = self.value
        else:
            value = validate_value(value)
        
        new = object.__new__(type(self))
        new.inline = inline
        new.name = name
        new.value = value
        return new
    
    
    @copy_docs(EmbedFieldBase.iter_contents)
    def iter_contents(self):
        name = self.name
        if (name is not None):
            yield name
        
        value = self.value
        if (value is not None):
            yield value



__all__ = ()

from scarletio import include

from ...application import SKU
from ...channel import Channel, ChannelType
from ...color import Color
from ...field_parsers import (
    bool_parser_factory, default_entity_parser_factory, entity_id_parser_factory, force_string_parser_factory,
    int_parser_factory, negated_bool_parser_factory, nullable_entity_parser_factory, nullable_flag_parser_factory,
    nullable_object_array_parser_factory, nullable_string_parser_factory, preinstanced_array_parser_factory,
    preinstanced_parser_factory
)
from ...field_putters import (
    bool_optional_putter_factory, entity_id_optional_putter_factory, entity_putter_factory, force_string_putter_factory,
    int_optional_putter_factory, int_putter_factory, negated_bool_optional_putter_factory,
    nullable_entity_array_putter_factory, nullable_entity_optional_putter_factory,
    nullable_flag_optional_putter_factory, nullable_string_optional_putter_factory, nullable_string_putter_factory,
    preinstanced_array_putter_factory, preinstanced_putter_factory, url_optional_putter_factory
)
from ...field_validators import (
    bool_validator_factory, entity_id_validator_factory, force_string_validator_factory,
    int_conditional_validator_factory, nullable_flag_validator_factory, nullable_object_array_validator_factory,
    nullable_string_validator_factory, preinstanced_array_validator_factory, preinstanced_validator_factory,
    url_optional_validator_factory
)
from ...role import Role
from ...user import ClientUserBase

from ..entity_select_default_value import EntitySelectDefaultValue, EntitySelectDefaultValueType
from ..media_info import MediaInfo
from ..media_item import MediaItem
from ..shared_fields import (
    parse_components, parse_custom_id, parse_emoji, put_components, put_custom_id, put_emoji,
    validate_components_factory, validate_custom_id, validate_emoji
)
from ..string_select_option import StringSelectOption

from .constants import (
    BUTTON_STYLE_DEFAULT, CONTENT_LENGTH_MAX, CONTENT_LENGTH_MIN, DESCRIPTION_LENGTH_MAX, DESCRIPTION_LENGTH_MIN,
    LABEL_LENGTH_MAX__BUTTON, LABEL_LENGTH_MAX__LABEL, MAX_LENGTH_DEFAULT, MAX_LENGTH_MAX, MAX_LENGTH_MIN,
    MAX_VALUES_DEFAULT, MAX_VALUES_MAX__ATTACHMENT_INPUT, MAX_VALUES_MAX__SELECT, MAX_VALUES_MIN, MIN_LENGTH_DEFAULT,
    MIN_LENGTH_MAX, MIN_LENGTH_MIN, MIN_VALUES_DEFAULT, MIN_VALUES_MAX__ATTACHMENT_INPUT, MIN_VALUES_MAX__SELECT,
    MIN_VALUES_MIN, PLACEHOLDER_LENGTH_MAX, SEPARATOR_SPACING_SIZE_DEFAULT, TEXT_INPUT_STYLE_DEFAULT, URL_LENGTH_MAX,
    VALUE_LENGTH_MAX
)
from .preinstanced import ButtonStyle, SeparatorSpacingSize, TextInputStyle


Component = include('Component')


# button_style

parse_button_style = preinstanced_parser_factory('style', ButtonStyle, BUTTON_STYLE_DEFAULT)
put_button_style = preinstanced_putter_factory('style')
validate_button_style = preinstanced_validator_factory('button_style', ButtonStyle)


# channel_types

parse_channel_types = preinstanced_array_parser_factory('channel_types', ChannelType)
put_channel_types = preinstanced_array_putter_factory('channel_types')
validate_channel_types = preinstanced_array_validator_factory('channel_types', ChannelType)


# color

parse_color = nullable_flag_parser_factory('accent_color', Color)
put_color = nullable_flag_optional_putter_factory('accent_color')
validate_color = nullable_flag_validator_factory('color', Color)


# components

parse_component__label = nullable_entity_parser_factory('component', NotImplemented, include = 'Component')
put_component__label = nullable_entity_optional_putter_factory(
    'component', NotImplemented, can_include_internals = True
)

def validate_component__label(component):
    """
    Validates whether the given label component is correct type.
    
    Parameters
    ----------
    component : ``None | Component``
        The component to validate.
    
    Returns
    -------
    component : ``None | Component``
    
    Raises
    ------
    TypeError
        - If `component` is given as an invalid type.
    """
    if component is None:
        return None
    
    if isinstance(component, Component):
        if component.type.layout_flags.nestable_into_label:
            return component
        
        raise ValueError(
            f'Component of type: {component.type.name} cannot be used as a label component, got {component!r}.'
        )
    
    raise TypeError(
        f'`component` can be `{Component.__name__}`, got {type(component).__name__}, {component!r}.'
    )


# components

validate_components__row = validate_components_factory(
    'row',
    'message / form',
    (
        'allowed_in_message',
        'allowed_in_form',
    ),
    (
        'nestable_into_row',
    ),
)

validate_components__container = validate_components_factory(
    'container',
    'message',
    (
        'allowed_in_message',
    ),
    (
        'nestable_into_container',
    ),
)

validate_components__section = validate_components_factory(
    'section',
    'message',
    (
        'allowed_in_message',
    ),
    (
        'nestable_into_section',
    ),
)


# content

parse_content = nullable_string_parser_factory('content')
put_content = nullable_string_putter_factory('content')
validate_content = nullable_string_validator_factory('content', CONTENT_LENGTH_MIN, CONTENT_LENGTH_MAX)


# default_values

parse_default_values = nullable_object_array_parser_factory('default_values', EntitySelectDefaultValue)
put_default_values = nullable_entity_array_putter_factory('default_values', EntitySelectDefaultValue)


def _validate_default_value_channel(channel):
    """
    Creates a default option from a channel.
    
    Parameters
    ----------
    channel : ``Channel``
        Channel to create default option from.
    
    Returns
    -------
    default_value : ``EntitySelectDefaultValue``
    """
    return EntitySelectDefaultValue.from_fields(EntitySelectDefaultValueType.channel, channel.id)


def _validate_default_value_role(role):
    """
    Creates a default option from a role.
    
    Parameters
    ----------
    role : ``Role``
        Role to create default option from.
    
    Returns
    -------
    default_value : ``EntitySelectDefaultValue``
    """
    return EntitySelectDefaultValue.from_fields(EntitySelectDefaultValueType.role, role.id)


def _validate_default_value_user(user):
    """
    Creates a default option from a user.
    
    Parameters
    ----------
    user : ``ClientUserBase``
        User to create default option from.
    
    Returns
    -------
    default_value : ``EntitySelectDefaultValue``
    """
    return EntitySelectDefaultValue.from_fields(EntitySelectDefaultValueType.user, user.id)


def _validate_default_value_option(default_value):
    """
    Creates a default option from a default option
    
    Parameters
    ----------
    default_value : ``EntitySelectDefaultValue``
        Just returns it.
    
    Returns
    -------
    default_value : ``EntitySelectDefaultValue``
    """
    return default_value


def _validate_default_value_tuple(default_value_tuple):
    """
    Creates a default option from a tuple.
    
    Parameters
    ----------
    default_value_tuple : `tuple`
        Tuple to create default option from.
    
    Returns
    -------
    default_value : ``EntitySelectDefaultValue``
    
    Raises
    ------
    ValueError
        - If tuple length is not `2`.
    """
    default_value_tuple_length = len(default_value_tuple)
    if default_value_tuple_length != 2:
        raise ValueError(
            f'When a default option is given as a `tuple` its length can be `2`, got {default_value_tuple_length!r}; '
            f'default_value = {default_value_tuple!r}.'
        )
    
    return EntitySelectDefaultValue(*default_value_tuple)


DEFAULT_OPTION_VALIDATORS_BY_TYPE = {
    Channel: _validate_default_value_channel,
    Role: _validate_default_value_role,
    ClientUserBase: _validate_default_value_user,
    EntitySelectDefaultValue: _validate_default_value_option,
    tuple: _validate_default_value_tuple,
}


def _validate_default_value_single(default_value):
    """
    Validates a single option.
    
    Parameters
    ----------
    default_value : `Channel | Role | ClientUserBase | EntitySelectDefaultValue | tuple`
        Default option to validate.
    
    Returns
    -------
    default_value : ``EntitySelectDefaultValue``
    
    Raises
    ------
    TypeError
        If a parameter has invalid type.
    ValueError
        If a parameter has invalid value.
    """
    default_value_type = type(default_value)
    
    try:
        validator = DEFAULT_OPTION_VALIDATORS_BY_TYPE[default_value_type]
    except KeyError:
        for validated_type, validator in DEFAULT_OPTION_VALIDATORS_BY_TYPE.items():
            if issubclass(default_value_type, validated_type):
                break
        
        else:
            raise TypeError(
                f'Default option can be `{Channel.__name__}`, `{Role.__name__}`, `{ClientUserBase.__name__}`,'
                f'`{EntitySelectDefaultValue.__name__}`, `tuple`, '
                f'got {default_value_type.__name__}; {default_value!r}.'
            ) from None
        
        
        DEFAULT_OPTION_VALIDATORS_BY_TYPE[default_value_type] = validator
    
    return validator(default_value)


def validate_default_values(default_values):
    """
    Validates the given default options.
    
    Parameters
    ----------
    default_values : ``None | iterable<Channel> | iterable<Role> | iterable<ClientUserBase> | iterable<EntitySelectDefaultValue> | iterable<(str | EntitySelectDefaultValueTyp, int | str)>>``
        Default options to validate.
    
    Returns
    -------
    default_values : `None | tuple<EntitySelectDefaultValue>`
    
    Raises
    ------
    TypeError
        If a parameter has invalid type.
    ValueError
        If a parameter has invalid value.
    """
    default_values_validated = None
    
    if default_values is None:
        return default_values_validated
    
    if getattr(type(default_values), '__iter__', None) is not None:
        for default_value in default_values:
            try:
                default_value = _validate_default_value_single(default_value)
            except (ValueError, TypeError) as err:
                err.args = (*err.args, f'default_values = {default_values!r}')
                raise
            
            if default_values_validated is None:
                default_values_validated = []
            default_values_validated.append(default_value)
        
        if (default_values_validated is not None):
            default_values_validated = (*default_values_validated,)
        
        return default_values_validated
        
    raise TypeError(
        f'`default_values` can be `None` or `iterable` of any of the following: `{Channel.__name__}`, '
        f'`{Role.__name__}`, `{ClientUserBase.__name__}`, `{EntitySelectDefaultValue.__name__}`; '
        f'got {type(default_values).__name__}; {default_values!r}.'
    )


# description

parse_description = nullable_string_parser_factory('description')
put_description = nullable_string_optional_putter_factory('description')
validate_description = nullable_string_validator_factory(
    'description', DESCRIPTION_LENGTH_MIN, DESCRIPTION_LENGTH_MAX
)


# divider

parse_divider = bool_parser_factory('divider', True)
put_divider = bool_optional_putter_factory('divider', True)
validate_divider = bool_validator_factory('divider', True)


# enabled

parse_enabled = negated_bool_parser_factory('disabled', True)
put_enabled = negated_bool_optional_putter_factory('disabled', True)
validate_enabled = bool_validator_factory('enabled', True)


# items

parse_items = nullable_object_array_parser_factory('items', MediaItem)
put_items = nullable_entity_array_putter_factory('items', MediaItem)


def validate_items(items):
    """
    Validates whether the given items are is correct type.
    
    Parameters
    ----------
    items : ``None | iterable<str> | iterable<MediaInfo>``
        The items to validate.
    
    Returns
    -------
    items : ``None | tuple<MediaInfo>``
    
    Raises
    ------
    TypeError
        - If `items`'s type is incorrect.
    """
    if items is None:
        return None
    
    if (getattr(items, '__iter__', None) is None):
        raise TypeError(
            f'`items` can be `None | iterable<str> | iterable<{MediaItem.__name__}>`, got '
            f'{type(items).__name__}; {items!r}.'
        )
    
    items_processed = None
    
    for item in items:
        if isinstance(item, str):
            item = MediaItem(MediaInfo(item))
        
        elif isinstance(item, MediaItem):
            pass
        
        else:
            raise TypeError(
                f'`items` can contain `str | {MediaItem.__name__}` elements, got '
                f'{type(item).__name__}; {item!r}; items = {items!r}.'
            )
        
        if (items_processed is None):
            items_processed = []
        
        items_processed.append(item)
    
    if (items_processed is not None):
        items_processed = tuple(items_processed)
    
    return items_processed


# label

parse_label = nullable_string_parser_factory('label')
put_label = nullable_string_optional_putter_factory('label')
validate_label = nullable_string_validator_factory('label', 0, max(LABEL_LENGTH_MAX__BUTTON, LABEL_LENGTH_MAX__LABEL))
validate_label__button = nullable_string_validator_factory('label', 0, LABEL_LENGTH_MAX__BUTTON)
validate_label__label = nullable_string_validator_factory('label', 0, LABEL_LENGTH_MAX__LABEL)


# max_length

parse_max_length = int_parser_factory('max_length', MAX_LENGTH_DEFAULT)
put_max_length = int_optional_putter_factory('max_length', MAX_LENGTH_DEFAULT)
validate_max_length = int_conditional_validator_factory(
    'max_length',
    MAX_LENGTH_MIN,
    (lambda max_length: max_length >= MAX_LENGTH_MIN and max_length <= MAX_LENGTH_MAX),
    f'>= {MAX_LENGTH_MIN} and <= {MAX_LENGTH_MAX},'
)


# max_values

parse_max_values = int_parser_factory('max_values', MAX_VALUES_DEFAULT)
put_max_values = int_optional_putter_factory('max_values', MAX_VALUES_DEFAULT)
validate_max_values = int_conditional_validator_factory(
    'max_values',
    MAX_VALUES_MIN,
    (
        lambda max_values: (
            max_values >= MAX_VALUES_MIN and
            max_values <= max(MAX_VALUES_MAX__ATTACHMENT_INPUT, MAX_VALUES_MAX__SELECT)
        )
    ),
    f'>= {MAX_VALUES_MIN} and <= {max(MAX_VALUES_MAX__ATTACHMENT_INPUT, MAX_VALUES_MAX__SELECT)}'
)
validate_max_values__attachment_input = int_conditional_validator_factory(
    'max_values',
    MAX_VALUES_MIN,
    (lambda max_values: max_values >= MAX_VALUES_MIN and max_values <= MAX_VALUES_MAX__ATTACHMENT_INPUT),
    f'>= {MAX_VALUES_MIN} and <= {MAX_VALUES_MAX__ATTACHMENT_INPUT},'
)
validate_max_values__select = int_conditional_validator_factory(
    'max_values',
    MAX_VALUES_MIN,
    (lambda max_values: max_values >= MAX_VALUES_MIN and max_values <= MAX_VALUES_MAX__SELECT),
    f'>= {MAX_VALUES_MIN} and <= {MAX_VALUES_MAX__SELECT},'
)

# media

parse_media__attachment_only = default_entity_parser_factory(
    'file', MediaInfo, default_factory = MediaInfo._create_empty
)
put_media__attachment_only = entity_putter_factory('file', MediaInfo)

parse_media = default_entity_parser_factory('media', MediaInfo, default_factory = MediaInfo._create_empty)
put_media = entity_putter_factory('media', MediaInfo)

def validate_media(media):
    """
    Validates whether the given media is correct type.
    
    Parameters
    ----------
    media : ``str | MediaInfo``
        The media to validate.
    
    Returns
    -------
    media : ``MediaInfo``
    
    Raises
    ------
    TypeError
        - If `media` is given as an invalid type.
    """
    if isinstance(media, str):
        return MediaInfo(media)
    
    if isinstance(media, MediaInfo):
        return media
    
    raise TypeError(
        f'`media` can be `str | {MediaInfo.__name__}`, got {type(media).__name__}, {media!r}.'
    )

# min_length

parse_min_length = int_parser_factory('min_length', MIN_LENGTH_DEFAULT)
put_min_length = int_optional_putter_factory('min_length', MIN_LENGTH_DEFAULT)
validate_min_length = int_conditional_validator_factory(
    'min_length',
    MIN_LENGTH_MIN,
    (lambda min_length: min_length >= MIN_LENGTH_MIN and min_length <= MIN_LENGTH_MAX),
    f'>= {MIN_LENGTH_MIN} and <= {MIN_LENGTH_MAX},'
)

# min_values

parse_min_values = int_parser_factory('min_values', MIN_VALUES_DEFAULT)
put_min_values = int_optional_putter_factory('min_values', MIN_VALUES_DEFAULT)
validate_min_values = int_conditional_validator_factory(
    'min_values',
    MIN_VALUES_MIN,
    (
        lambda min_values: (
            min_values >= MIN_VALUES_MIN and
            min_values <= max(MIN_VALUES_MAX__ATTACHMENT_INPUT, MIN_VALUES_MAX__SELECT)
        )
    ),
    f'>= {MIN_VALUES_MIN} and <= {max(MIN_VALUES_MAX__ATTACHMENT_INPUT, MIN_VALUES_MAX__SELECT)}'
)
validate_min_values__attachment_input = int_conditional_validator_factory(
    'min_values',
    MIN_VALUES_MIN,
    (lambda min_values: min_values >= MIN_VALUES_MIN and min_values <= MIN_VALUES_MAX__ATTACHMENT_INPUT),
    f'>= {MIN_VALUES_MIN} and <= {MIN_VALUES_MAX__ATTACHMENT_INPUT},'
)
validate_min_values__select = int_conditional_validator_factory(
    'min_values',
    MIN_VALUES_MIN,
    (lambda min_values: min_values >= MIN_VALUES_MIN and min_values <= MIN_VALUES_MAX__SELECT),
    f'>= {MIN_VALUES_MIN} and <= {MIN_VALUES_MAX__SELECT},'
)


# name

parse_name = force_string_parser_factory('name')
put_name = force_string_putter_factory('name')
validate_name = force_string_validator_factory('name', 0, 1024)


# options

parse_options = nullable_object_array_parser_factory('options', StringSelectOption)
put_options = nullable_entity_array_putter_factory('options', StringSelectOption)
validate_options = nullable_object_array_validator_factory('options', StringSelectOption)

# placeholder

parse_placeholder = nullable_string_parser_factory('placeholder')
put_placeholder = nullable_string_optional_putter_factory('placeholder')
validate_placeholder = nullable_string_validator_factory('placeholder', 0, PLACEHOLDER_LENGTH_MAX)

# required

parse_required = bool_parser_factory('required', True)
put_required = bool_optional_putter_factory('required', True)
validate_required = bool_validator_factory('required', True)


# size

parse_size = int_parser_factory('size', 0)
put_size = int_putter_factory('size')
validate_size = int_conditional_validator_factory(
    'size',
    0,
    lambda size : size >= 0,
    '>= 0',
)

# sku_id

parse_sku_id = entity_id_parser_factory('sku_id')
put_sku_id = entity_id_optional_putter_factory('sku_id')
validate_sku_id = entity_id_validator_factory('sku_id', SKU)


# spacing_size

parse_spacing_size = preinstanced_parser_factory('spacing', SeparatorSpacingSize, SEPARATOR_SPACING_SIZE_DEFAULT)
put_spacing_size = preinstanced_putter_factory('spacing')
validate_spacing_size = preinstanced_validator_factory('spacing_size', SeparatorSpacingSize)


# spoiler

parse_spoiler = bool_parser_factory('spoiler', False)
put_spoiler = bool_optional_putter_factory('spoiler', False)
validate_spoiler = bool_validator_factory('spoiler', False)


# text_input_style

parse_text_input_style = preinstanced_parser_factory('style', TextInputStyle, TEXT_INPUT_STYLE_DEFAULT)
put_text_input_style = preinstanced_putter_factory('style')
validate_text_input_style = preinstanced_validator_factory('text_input_style', TextInputStyle)


# thumbnail

parse_thumbnail = nullable_entity_parser_factory('accessory', NotImplemented, include = 'Component')
put_thumbnail = nullable_entity_optional_putter_factory('accessory', NotImplemented, can_include_internals = True)

def validate_thumbnail(thumbnail):
    """
    Validates whether the given thumbnail is correct type.
    
    Parameters
    ----------
    thumbnail : ``None | Component``
        The thumbnail to validate.
    
    Returns
    -------
    thumbnail : ``None | Component``
    
    Raises
    ------
    TypeError
        - If `thumbnail` is given as an invalid type.
    """
    if thumbnail is None:
        return None
    
    if isinstance(thumbnail, Component):
        if thumbnail.type.layout_flags.section_thumbnail:
            return thumbnail
        
        raise ValueError(
            f'Component of type: {thumbnail.type.name} cannot be used as a section thumbnail, got {thumbnail!r}.'
        )
    
    raise TypeError(
        f'`thumbnail` can be `{Component.__name__}`, got {type(thumbnail).__name__}, {thumbnail!r}.'
    )


# url

parse_url = nullable_string_parser_factory('url')
put_url = url_optional_putter_factory('url')
validate_url = url_optional_validator_factory('url', length_max = URL_LENGTH_MAX)


# value

parse_value = nullable_string_parser_factory('value')
put_value = nullable_string_optional_putter_factory('value')
validate_value = nullable_string_validator_factory('value', 0, VALUE_LENGTH_MAX)



__all__ = ('ChannelMetadataGuildBase',)

from scarletio import copy_docs, export, include

from ...permission.permission import PERMISSION_MASK_VIEW_CHANNEL

from .fields import parse_name, parse_parent_id, put_name, put_parent_id, validate_name, validate_parent_id

from .base import ChannelMetadataBase


Client = include('Client')


@export
class ChannelMetadataGuildBase(ChannelMetadataBase):
    """
    Base guild channel metadata type.
    
    Attributes
    ----------
    _cache_permission : ``None | dict<int, Permission>``
        A `user_id` to ``Permission`` relation mapping for caching permissions. Defaults to `None`.
    name : `str`
        The channel's name.
    parent_id : `int`
        The channel's parent's identifier.
    
    Class Attributes
    ----------------
    order_group: `int` = `0`
        The channel's order group used when sorting channels.
    """
    __slots__ = ('_cache_permission', 'name', 'parent_id')
    
    
    def __new__(
        cls,
        *,
        name = ...,
        parent_id = ...,
    ):
        """
        Creates a new guild base channel metadata from the given parameters.
        
        Parameters
        ----------
        name : `str`, Optional (Keyword only)
            The channel's name.
        parent_id : ``None | int | Channel``, Optional (Keyword only)
            The channel's parent's identifier.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # name
        if name is ...:
            name = ''
        else:
            name = validate_name(name)
        
        # parent_id
        if parent_id is ...:
            parent_id = 0
        else:
            parent_id = validate_parent_id(parent_id)
        
        # Construct
        self = object.__new__(cls)
        self._cache_permission = None
        self.name = name
        self.parent_id = parent_id
        return self
    
    
    @classmethod
    @copy_docs(ChannelMetadataBase.from_keyword_parameters)
    def from_keyword_parameters(cls, keyword_parameters):
        return cls(
            name = keyword_parameters.pop('name', ...),
            parent_id = keyword_parameters.pop('parent_id', ...),
        )
    
    
    @copy_docs(ChannelMetadataBase.__hash__)
    def __hash__(self):
        hash_value = ChannelMetadataBase.__hash__(self)
        
        # name
        name = self.name
        if name:
            hash_value ^= hash(name)
        
        # parent_id
        hash_value ^= self.parent_id
        
        return hash_value
    
    
    @classmethod
    @copy_docs(ChannelMetadataBase.from_data)
    def from_data(cls, data):
        self = super(ChannelMetadataGuildBase, cls).from_data(data)
        
        self._cache_permission = None
        
        return self
    
    
    @copy_docs(ChannelMetadataBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        
        # name
        if self.name != other.name:
            return False
        
        # parent_id
        if self.parent_id != other.parent_id:
            return False
        
        return True
    
    
    @copy_docs(ChannelMetadataBase.to_data)
    def to_data(self, *, defaults = False, include_internals = False):
        data = ChannelMetadataBase.to_data(self, defaults = defaults, include_internals = include_internals)
        
        # name
        put_name(self.name, data, defaults)
        
        # parent_id
        put_parent_id(self.parent_id, data, defaults)
        
        return data
    
    
    @classmethod
    @copy_docs(ChannelMetadataBase._from_partial_data)
    def _from_partial_data(cls, data):
        self = super(ChannelMetadataGuildBase, cls)._from_partial_data(data)
        
        if (data is not None):
            name = data.get('name', None)
            if (name is not None):
                self.name = name
        
        return self
    
    
    @copy_docs(ChannelMetadataBase._update_attributes)
    def _update_attributes(self, data):
        ChannelMetadataBase._update_attributes(self, data)
        
        self._cache_permission = None
        
        # name
        self.name = parse_name(data)
        
        # parent_id
        self.parent_id = parse_parent_id(data)
        

    @copy_docs(ChannelMetadataBase._difference_update_attributes)
    def _difference_update_attributes(self, data):
        old_attributes = ChannelMetadataBase._difference_update_attributes(self, data)
        
        self._cache_permission = None
        
        # name
        name = parse_name(data)
        if self.name != name:
            old_attributes['name'] = self.name
            self.name = name
        
        # parent_id
        parent_id = parse_parent_id(data)
        if self.parent_id != parent_id:
            old_attributes['parent_id'] = self.parent_id
            self.parent_id = parent_id
        
        return old_attributes
    
    
    @classmethod
    @copy_docs(ChannelMetadataBase._create_empty)
    def _create_empty(cls):
        self = super(ChannelMetadataGuildBase, cls)._create_empty()
        
        self._cache_permission = None
        self.parent_id = 0
        self.name = ''
        
        return self
    
    
    @copy_docs(ChannelMetadataBase.copy)
    def copy(self):
        new = object.__new__(type(self))
        new._cache_permission = None
        new.parent_id = self.parent_id
        new.name = self.name
        return new
    
    
    def copy_with(
        self,
        *,
        name = ...,
        parent_id = ...,
    ):
        """
        Copies the guild base channel metadata with the given fields.
        
        Parameters
        ----------
        name : `str`, Optional (Keyword only)
            The channel's name.
        parent_id : ``None | int | Channel``, Optional (Keyword only)
            The channel's parent's identifier.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # name
        if name is ...:
            name = self.name
        else:
            name = validate_name(name)
        
        # parent_id
        if parent_id is ...:
            parent_id = self.parent_id
        else:
            parent_id = validate_parent_id(parent_id)
        
        # Construct
        self = object.__new__(type(self))
        self._cache_permission = None
        self.name = name
        self.parent_id = parent_id
        return self
    
    
    @copy_docs(ChannelMetadataBase.copy_with_keyword_parameters)
    def copy_with_keyword_parameters(self, keyword_parameters):
        return self.copy_with(
            name = keyword_parameters.pop('name', ...),
            parent_id = keyword_parameters.pop('parent_id', ...),
        )
    
    
    @copy_docs(ChannelMetadataBase._get_users)
    def _get_users(self, channel_entity):
        return [*self._iter_users(channel_entity)]
    
    
    @copy_docs(ChannelMetadataBase._iter_users)
    def _iter_users(self, channel_entity):
        guild = channel_entity.guild
        if (guild is not None):
            for user in guild.users.values():
                if self._get_permissions_for(channel_entity, user) & PERMISSION_MASK_VIEW_CHANNEL:
                    yield user
    
    
    @copy_docs(ChannelMetadataBase._iter_clients)
    def _iter_clients(self, channel_entity):
        guild = channel_entity.guild
        if (guild is not None):
            for client in guild.clients:
                if self._get_cached_permissions_for(channel_entity, client) & PERMISSION_MASK_VIEW_CHANNEL:
                    yield client
    
    
    @copy_docs(ChannelMetadataBase._get_partial)
    def _get_partial(self, channel_entity):
        guild = channel_entity.guild
        if guild is None:
            return True
        
        return False if guild.clients else True
    
    
    @copy_docs(ChannelMetadataBase._get_cached_permissions_for)
    def _get_cached_permissions_for(self, channel_entity, user):
        if not isinstance(user, Client):
            return self._get_permissions_for(channel_entity, user)
        
        permission_cache = self._cache_permission
        if permission_cache is None:
            self._cache_permission = permission_cache = {}
        else:
            try:
                return permission_cache[user.id]
            except KeyError:
                pass
        
        permissions = self._get_permissions_for(channel_entity, user)
        permission_cache[user.id] = permissions
        return permissions
    
    
    @copy_docs(ChannelMetadataBase._invalidate_cache_permission)
    def _invalidate_cache_permission(self):
        self._cache_permission = None



__all__ = ()

from ...channel import Channel
from ...emoji import Emoji, create_partial_emoji_data, create_partial_emoji_from_data
from ...field_parsers import (
    entity_id_array_parser_factory, entity_id_parser_factory, force_string_parser_factory,
    nullable_functional_parser_factory, nullable_string_parser_factory
)
from ...field_putters import (
    entity_id_putter_factory, force_string_putter_factory, optional_entity_id_array_optional_putter_factory,
    nullable_functional_optional_putter_factory, nullable_string_putter_factory
)
from ...field_validators import (
    entity_id_array_validator_factory, entity_id_validator_factory, force_string_validator_factory,
    nullable_entity_validator_factory, nullable_string_validator_factory
)
from ...role import Role

# channel_ids

parse_channel_ids = entity_id_array_parser_factory('channel_ids')
put_channel_ids = optional_entity_id_array_optional_putter_factory('channel_ids')
validate_channel_ids = entity_id_array_validator_factory('channel_ids', Channel)

# description

parse_description = nullable_string_parser_factory('description')
put_description = nullable_string_putter_factory('description')
validate_description = nullable_string_validator_factory('description', 0, 1024)

# emoji

parse_emoji = nullable_functional_parser_factory('emoji', create_partial_emoji_from_data)
put_emoji = nullable_functional_optional_putter_factory('emoji', create_partial_emoji_data)
validate_emoji = nullable_entity_validator_factory('emoji', Emoji)

# id

parse_id = entity_id_parser_factory('id')
put_id = entity_id_putter_factory('id')
validate_id = entity_id_validator_factory('option_id')

# name

parse_name = force_string_parser_factory('title')
put_name = force_string_putter_factory('title')
validate_name = force_string_validator_factory('name', 0, 1024)

# role_ids

parse_role_ids = entity_id_array_parser_factory('role_ids')
put_role_ids = optional_entity_id_array_optional_putter_factory('role_ids')
validate_role_ids = entity_id_array_validator_factory('role_ids', Role)



__all__ = ('ScheduledEventUnsubscribeEvent',)

from ..scheduled_event_subscribe_event import ScheduledEventSubscribeEvent


class ScheduledEventUnsubscribeEvent(ScheduledEventSubscribeEvent):
    """
    Represents a `GUILD_SCHEDULED_EVENT_USER_REMOVE` event.
    
    Attributes
    ----------
    guild_id : `int`
        The guild's identifier where the event is for.
    scheduled_event_id : `int`
        The scheduled event's identifier.
    user_id : `int`
        The identifier of the user, who unsubscribed to the event.
    """
    __slots__ = ()



from .audio_settings import *
from .constants import *
from .fields import *


__all__ = (
    'AUDIO_SETTINGS_DEFAULT',
    *audio_settings.__all__,
    *constants.__all__,
    *fields.__all__,
)


AUDIO_SETTINGS_DEFAULT = AudioSettings(
    channels = 2,
    frame_length = 20,
    sampling_rate = 48000,
)



__all__ = ()

DISCRIMINATOR_VALUE_MIN = 0
DISCRIMINATOR_VALUE_MAX = 9999

DISPLAY_NAME_LENGTH_MAX = 32

NAME_LENGTH_MIN = 2
NAME_LENGTH_MAX = 32

WEBHOOK_NAME_LENGTH_MIN = 1
WEBHOOK_NAME_LENGTH_MAX = 80



__all__ = ('SlashCommandCategory',)

from scarletio import RichAttributeErrorBaseType, copy_docs, include

from .....discord.application_command import ApplicationCommandOption, ApplicationCommandOptionType
from .....discord.application_command.application_command.constants import (
    OPTIONS_MAX as APPLICATION_COMMAND_OPTIONS_MAX
)
from .....discord.client import Client
from .....discord.interaction import InteractionEvent

from ...constants import (
    APPLICATION_COMMAND_CATEGORY_DEEPNESS_MAX, APPLICATION_COMMAND_OPTION_TYPE_SUB_COMMAND,
    APPLICATION_COMMAND_OPTION_TYPE_SUB_COMMAND_CATEGORY
)
from ...exceptions import SlashCommandParameterConversionError, handle_command_exception
from ...interfaces.autocomplete import AutocompleteInterface
from ...interfaces.exception_handler import ExceptionHandlerInterface
from ...interfaces.nestable import NestableInterface
from ...interfaces.self_reference import SelfReferenceInterface

from ..command_base import CommandBase

from .helpers import _reset_parent_schema
from .slash_command_parameter_auto_completer import SlashCommandParameterAutoCompleter


Slasher = include('Slasher')
SlashCommand = include('SlashCommand')


class SlashCommandCategory(
    AutocompleteInterface,
    ExceptionHandlerInterface,
    NestableInterface,
    SelfReferenceInterface,
    RichAttributeErrorBaseType,
):
    """
    Represents an application command's backend implementation.
    
    Attributes
    ----------
    _auto_completers : `None | list<SlashCommandParameterAutoCompleter>`
        Auto completer functions by.
    
    _deepness : `int`
        How nested the category is.
    
    _exception_handlers : `None | list<CoroutineFunction>`
        Exception handlers added with ``.error`` to the interaction handler.
    
    _parent_reference : `None | WeakReferer<SelfReferenceInterface>`
        The parent slash command of the category if any.
    
    _self_reference : `None | WeakReferer<instance>`
        Back reference to the slasher application command category.
        
        Used by sub commands to access the parent entity.
    
    _sub_commands : `None | dict<str, SlashCommandFunction | SlashCommandCategory>`
        The sub-commands of the category.
    
    default : `bool`
        Whether the command is the default command in it's category.
    
    description : `str`
        The slash command's description.
    
    name : `str`
        The name of the slash sub-category.
    """
    __slots__ = (
        '__weakref__', '_auto_completers', '_deepness', '_exception_handlers', '_parent_reference',  '_self_reference',
        '_sub_commands', 'default', 'description', 'name'
    )
    
    def __new__(cls, name, description, default, deepness):
        """
        Creates a new slash command category.
        
        Parameters
        ----------
        name : `str`
            The name of the slash command.
        
        description : `str`
            The slash command's description.
        
        default : `bool`
            Whether the command is the default command in it's category.
        
        deepness : `int`
            How nested the category is.
        
        Raises
        ------
        RuntimeError
            Maximum deepness reached.
        """
        if deepness > APPLICATION_COMMAND_CATEGORY_DEEPNESS_MAX:
            raise RuntimeError('Maximum deepness reached. Cannot add anymore sub-category under sub-categories.')
        
        self = object.__new__(cls)
        self._auto_completers = None
        self._deepness = deepness
        self._exception_handlers = None
        self._parent_reference = None
        self._self_reference = None
        self._sub_commands = None
        self.default = default
        self.description = description
        self.name = name
        
        return self
    
    
    def __repr__(self):
        """Returns the slash command category's representation."""
        repr_parts = ['<', type(self).__name__]
        
        # name
        repr_parts.append(' name = ')
        repr_parts.append(repr(self.name))
        
        # description
        repr_parts.append(', description = ')
        repr_parts.append(repr(self.description))
        
        # default
        default = self.default
        if default:
            repr_parts.append(', default = ')
            repr_parts.append(repr(default))
        
        # _auto_completers
        auto_completers = self._auto_completers
        if (auto_completers is not None):
            repr_parts.append(', auto_completers = ')
            repr_parts.append(repr(auto_completers))
        
        # _exception_handlers
        exception_handlers = self._exception_handlers
        if (exception_handlers is not None):
            repr_parts.append(', exception_handlers = ')
            repr_parts.append(repr(exception_handlers))
        
        # _sub_commands
        sub_commands = self._sub_commands
        if (sub_commands is not None):
            repr_parts.append(', sub_commands = ')
            repr_parts.append(repr(sub_commands))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns the slash command category's hash value."""
        hash_value = 0
        
        # _auto_completers
        auto_completers = self._auto_completers
        if (auto_completers is not None):
            hash_value ^= len(auto_completers)
            
            for auto_completer in auto_completers:
                hash_value ^= hash(auto_completer)
        
        # _deepness
        # Internal field
        
        # _exception_handlers
        exception_handlers = self._exception_handlers
        if (exception_handlers is not None):
            hash_value ^= len(exception_handlers) << 4
            
            for exception_handler in exception_handlers:
                try:
                    exception_handler_hash_value = hash(exception_handler)
                except TypeError:
                    exception_handler_hash_value = object.__hash__(exception_handler)
                hash_value ^= exception_handler_hash_value
        
        # _self_reference
        # Internal field
        
        # _sub_commands
        sub_commands = self._sub_commands
        if (sub_commands is not None):
            hash_value ^= len(sub_commands) << 8
            
            for sub_command in sub_commands:
                hash_value ^= hash(sub_command)
        
        # _parent_reference
        # Internal field
        
        # default
        hash_value ^= self.default << 12
        
        # description
        description = self.description
        hash_value ^= hash(description)
        
        # name
        name = self.name
        if name != description:
            hash_value ^= hash(name)
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two slash commands categories are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # _auto_completers
        if self._auto_completers != other._auto_completers:
            return False
        
        # _deepness
        # Internal Field
        
        # _exception_handlers
        if self._exception_handlers != other._exception_handlers:
            return False
        
        # _self_reference
        # Internal Field
        
        # _sub_commands
        if self._sub_commands != other._sub_commands:
            return False
        
        # _parent_reference
        # Internal Field
        
        # default
        if self.default != other.default:
            return False
        
        # description
        if self.description != other.description:
            return False
        
        # name
        if self.name != other.name:
            return False
        
        return True
    
    
    def __format__(self, code):
        """Formats the command in a format string."""
        if not code:
            return str(self)
        
        if code == 'm':
            return self.mention
        
        raise ValueError(
            f'Unknown format code {code!r} for {type(self).__name__}; {self!r}. '
            f'Available format codes: {""!r}, {"m"!r}.'
        )
    
    
    async def invoke(self, client, interaction_event, options):
        """
        Calls the slash command category.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The respective client who received the event.
        
        interaction_event : ``InteractionEvent``
            The received interaction event.
        
        options : `None | list<InteractionEventChoice>`
            Options bound to the category.
        """
        sub_commands = self._sub_commands
        if sub_commands is None:
            return
        
        if (options is None) or len(options) != 1:
            return
        
        option = options[0]
        
        try:
            sub_command = sub_commands[option.name]
        except KeyError:
            pass
        else:
            await sub_command.invoke(client, interaction_event, option.options)
            return
        
        # Do not put this into the `except` branch.
        await handle_command_exception(
            self,
            client,
            interaction_event,
            SlashCommandParameterConversionError(
                None,
                option.name,
                'sub-command',
                [*sub_commands.keys()],
            )
        )
        return
    
    
    def copy(self):
        """
        Copies the slash command category.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        
        # _auto_completers
        auto_completers = self._auto_completers
        if (auto_completers is not None):
            auto_completers = auto_completers.copy()
        new._auto_completers = auto_completers
        
        # _deepness
        new._deepness = self._deepness
        
        # _exception_handlers
        exception_handlers = self._exception_handlers
        if (exception_handlers is not None):
            exception_handlers = exception_handlers.copy()
        new._exception_handlers = exception_handlers
        
        # _self_reference
        new._self_reference = None
        
        # _sub_commands
        sub_commands = self._ub_commands
        if (sub_commands is not None):
            sub_commands = {name: category.copy() for name, category in sub_commands.items()}
        new._sub_commands = sub_commands
        
        # _parent_reference
        new._parent_reference = None
        
        # default
        new.default = self.default
        
        # description
        new.description = self.description
        
        # name
        new.name = self.name
        
        return new
    
    
    async def invoke_auto_completion(self, client, interaction_event, auto_complete_option):
        """
        Calls the respective auto completion function of the command.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The respective client who received the event.
        interaction_event : ``InteractionEvent``
            The received interaction event.
        auto_complete_option : `InteractionMetadataApplicationCommandAutocomplete | InteractionOption`
            The option to autocomplete.
        """
        auto_complete_option_type = auto_complete_option.type
        if (
            (auto_complete_option_type is APPLICATION_COMMAND_OPTION_TYPE_SUB_COMMAND) or
            (auto_complete_option_type is APPLICATION_COMMAND_OPTION_TYPE_SUB_COMMAND_CATEGORY)
        ):
            options = auto_complete_option.options
            if (options is not None):
                option = options[0]
                sub_commands = self._sub_commands
                if (sub_commands is not None):
                    try:
                        sub_command = sub_commands[option.name]
                    except KeyError:
                        pass
                    else:
                        await sub_command.invoke_auto_completion(client, interaction_event, option)
    
    
    @copy_docs(NestableInterface._make_command_instance_from_parameters)
    def _make_command_instance_from_parameters(self, function, positional_parameters, keyword_parameters):
        return SlashCommand(function, *positional_parameters, **keyword_parameters)
    
    
    @copy_docs(NestableInterface._store_command_instance)
    def _store_command_instance(self, command):
        if isinstance(command, SlashCommand):
            instance = self._add_application_command(command)
            return True, instance
        
        return False, None
    
    
    def _add_application_command(self, command):
        """
        Adds a sub-command or sub-category to the slash command.
        
        Parameters
        ----------
        command : ``SlashCommand``
            The slash command to add.
        
        Returns
        -------
        as_sub_command : `SlashCommandFunction | SlashCommandCategory`
            The command as sub-command.
        
        Raises
        ------
        RuntimeError
            - The ``SlashCommand`` reached the maximal amount of children.
            - Cannot add anymore sub-category under sub-categories.
            - If the command to add is a default sub-command meanwhile the category already has one.
        """
        sub_commands = self._sub_commands
        if (
            (sub_commands is not None) and
            (len(sub_commands) == APPLICATION_COMMAND_OPTIONS_MAX) and
            (command.name not in sub_commands)
        ):
            raise RuntimeError(
                f'The {type(self).__name__} reached the maximal amount of children '
                f'({APPLICATION_COMMAND_OPTIONS_MAX}).'
            )
        
        as_sub_command = command.as_sub_command(self._deepness + 1)
        
        if (sub_commands is not None) and command.default:
            for sub_command in sub_commands.values():
                if sub_command.default and (sub_command.name != command.name):
                    raise RuntimeError(
                        f'{self!r} already has a default command.'
                    )
        
        as_sub_command._parent_reference = self.get_self_reference()
        
        if sub_commands is None:
            sub_commands = {}
            self._sub_commands = sub_commands
        
        sub_commands[command.name] = as_sub_command
        
        _reset_parent_schema(self)
        
        # Resolve auto completers recursively
        parent = self
        while True:
            auto_completers = parent._auto_completers
            if (auto_completers is not None):
                for auto_completer in auto_completers:
                    as_sub_command._try_resolve_auto_completer(auto_completer)
            
            if isinstance(parent, Slasher):
                break
            
            parent_reference = parent._parent_reference
            if (parent_reference is None):
                break
            
            parent = parent_reference()
            if (parent is None):
                break
        
        return as_sub_command
    
    
    @copy_docs(AutocompleteInterface._register_auto_completer)
    def _register_auto_completer(self, function, parameter_names):
        auto_completer = self._make_auto_completer(function, parameter_names)
        self._store_auto_completer(auto_completer)
        
        resolved = 0
        sub_commands = self._sub_commands
        if (sub_commands is not None):
            for sub_command in sub_commands.values():
                resolved += sub_command._try_resolve_auto_completer(auto_completer)
        
        if resolved:
            _reset_parent_schema(self)
        
        
        return auto_completer
    
    
    def _try_resolve_auto_completer(self, auto_completer):
        """
        Tries to register auto completer to the slasher application command function.
        
        Parameters
        ----------
        auto_completer : ``SlashCommandParameterAutoCompleter``
            The auto completer.
        
        Returns
        -------
        resolved : `int`
            The amount of parameters resolved.
        """
        resolved = 0
        
        sub_commands = self._sub_commands
        if (sub_commands is not None):
            for sub_command in sub_commands.values():
                 resolved += sub_command._try_resolve_auto_completer(auto_completer)
        
        return resolved    
    
    
    def as_option(self):
        """
        Returns the slash command category as an application command option.
        
        Returns
        -------
        option : ``ApplicationCommandOption``
        """
        sub_commands = self._sub_commands
        if sub_commands is None:
            options = None
        else:
            options = [sub_command.as_option() for name, sub_command in sorted(sub_commands.items())]
        
        return ApplicationCommandOption(
            self.name,
            self.description,
            ApplicationCommandOptionType.sub_command_group,
            options = options,
            default = self.default,
        )
    
    
    # ---- Mention ----
    
    @property
    @copy_docs(CommandBase.mention)
    def mention(self):
        return self._mention_recursive()
    
    
    def _mention_recursive(self, *sub_command_names):
        """
        Returns the application command category's mention.
        
        Called by ``.mention`` to include the sub-commands' names.
        
        Parameters
        ----------
        *sub_command_names : `str`
            Already included sub-command name stack to mention.
        
        Returns
        -------
        mention : `str`
        """
        parent_reference = self._parent_reference
        if parent_reference is None:
            parent = None
        else:
            parent = parent_reference()
        
        if parent is None:
            return ''
        
        return parent._mention_recursive(self.name, *sub_command_names)
    
    
    @copy_docs(CommandBase.mention_at)
    def mention_at(self, guild):
        return self._mention_at_recursive(guild)
    
    
    def _mention_at_recursive(self, guild, *sub_command_names):
        """
        Returns the application command category's mention.
        
        Called by ``.mention`` to include the sub-commands' names.
        
        Parameters
        ----------
        guild : `Guild | int`
            The guild to mention the command at.
        
        *sub_command_names : `str`
            Already included sub-command name stack to mention.
        
        Returns
        -------
        mention : `str`
        """
        parent_reference = self._parent_reference
        if parent_reference is None:
            parent = None
        else:
            parent = parent_reference()
        
        if parent is None:
            return ''
        
        return parent._mention_at_recursive(guild, self.name, *sub_command_names)
